from __future__ import annotations
from anise import Almanac
from anise import analysis
from anise import rotation
from anise import time
import numpy
import typing

@typing.final
class AzElRange:
    """A structure that stores the result of Azimuth, Elevation, Range, Range rate calculation."""

    azimuth_deg: float
    elevation_deg: float
    epoch: time.Epoch
    light_time: time.Duration
    mask_deg: float
    obstructed_by: Frame
    range_km: float
    range_rate_km_s: float

    def __init__(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        """Initialize self.  See help(type(self)) for accurate signature."""

    def __new__(
        cls,
        epoch: time.Epoch,
        azimuth_deg: float,
        elevation_deg: float,
        range_km: float,
        range_rate_km_s: float,
        obstructed_by: typing.Optional[Frame] = None,
        mask_deg: typing.Optional[float] = None,
    ) -> AzElRange:
        """A structure that stores the result of Azimuth, Elevation, Range, Range rate calculation."""

    def elevation_above_mask_deg(self) -> float:
        """Returns the elevation above the terrain mask for this azimuth, in degrees.
        If the terrain mask was zero at this azimuth, then the elevation above mask is equal to the elevation_deg field."""

    def is_obstructed(self) -> bool:
        """Returns whether there is an obstruction."""

    def is_valid(self) -> bool:
        """Returns false if the range is less than one millimeter, or any of the angles are NaN."""

    def __eq__(self, value: typing.Any) -> bool:
        """Return self==value."""

    def __ge__(self, value: typing.Any) -> bool:
        """Return self>=value."""

    def __getnewargs__(self) -> typing.Tuple:
        """Allows for pickling the object"""

    def __gt__(self, value: typing.Any) -> bool:
        """Return self>value."""

    def __le__(self, value: typing.Any) -> bool:
        """Return self<=value."""

    def __lt__(self, value: typing.Any) -> bool:
        """Return self<value."""

    def __ne__(self, value: typing.Any) -> bool:
        """Return self!=value."""

    def __repr__(self) -> str:
        """Return repr(self)."""

    def __str__(self) -> str:
        """Return str(self)."""

@typing.final
class Covariance:
    matrix: numpy.ndarray

    def __init__(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        """Initialize self.  See help(type(self)) for accurate signature."""

    def __new__(cls, covar: typing.Any, local_frame: typing.Any): ...
    def __repr__(self) -> str:
        """Return repr(self)."""

    def __str__(self) -> str:
        """Return str(self)."""

@typing.final
class DataType:
    def __init__(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        """Initialize self.  See help(type(self)) for accurate signature."""

    def __new__(cls): ...
    def __eq__(self, value: typing.Any) -> bool:
        """Return self==value."""

    def __ge__(self, value: typing.Any) -> bool:
        """Return self>=value."""

    def __gt__(self, value: typing.Any) -> bool:
        """Return self>value."""

    def __int__(self) -> None:
        """int(self)"""

    def __le__(self, value: typing.Any) -> bool:
        """Return self<=value."""

    def __lt__(self, value: typing.Any) -> bool:
        """Return self<value."""

    def __ne__(self, value: typing.Any) -> bool:
        """Return self!=value."""

    def __repr__(self) -> str:
        """Return repr(self)."""
    Type10SpaceCommandTLE: DataType = ...
    Type12HermiteEqualStep: DataType = ...
    Type13HermiteUnequalStep: DataType = ...
    Type14ChebyshevUnequalStep: DataType = ...
    Type15PrecessingConics: DataType = ...
    Type17Equinoctial: DataType = ...
    Type18ESOCHermiteLagrange: DataType = ...
    Type19ESOCPiecewise: DataType = ...
    Type1ModifiedDifferenceArray: DataType = ...
    Type20ChebyshevDerivative: DataType = ...
    Type21ExtendedModifiedDifferenceArray: DataType = ...
    Type2ChebyshevTriplet: DataType = ...
    Type3ChebyshevSextuplet: DataType = ...
    Type5DiscreteStates: DataType = ...
    Type8LagrangeEqualStep: DataType = ...
    Type9LagrangeUnequalStep: DataType = ...

@typing.final
class DragData:
    area_m2: float
    coeff_drag: float

    def __init__(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        """Initialize self.  See help(type(self)) for accurate signature."""

    def __new__(cls, area_m2: typing.Any, coeff_drag: typing.Any = None): ...
    @staticmethod
    def from_asn1(data: bytes) -> DragData:
        """Decodes an ASN.1 DER encoded byte array into a DragData object."""

    def to_asn1(self) -> bytes:
        """Encodes this DragData object into an ASN.1 DER encoded byte array."""

    def __repr__(self) -> str:
        """Return repr(self)."""

    def __str__(self) -> str:
        """Return str(self)."""

@typing.final
class Ellipsoid:
    """Only the tri-axial Ellipsoid shape model is currently supported by ANISE.
    This is directly inspired from SPICE PCK.
    > For each body, three radii are listed: The first number is
    > the largest equatorial radius (the length of the semi-axis
    > containing the prime meridian), the second number is the smaller
    > equatorial radius, and the third is the polar radius.

    Example: Radii of the Earth.

    BODY399_RADII     = ( 6378.1366   6378.1366   6356.7519 )"""

    polar_radius_km: float
    semi_major_equatorial_radius_km: float
    semi_minor_equatorial_radius_km: float

    def __init__(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        """Initialize self.  See help(type(self)) for accurate signature."""

    def __new__(
        cls,
        semi_major_equatorial_radius_km: float,
        polar_radius_km: typing.Optional[float] = None,
        semi_minor_equatorial_radius_km: typing.Optional[float] = None,
    ) -> Ellipsoid:
        """Only the tri-axial Ellipsoid shape model is currently supported by ANISE.
        This is directly inspired from SPICE PCK.
        > For each body, three radii are listed: The first number is
        > the largest equatorial radius (the length of the semi-axis
        > containing the prime meridian), the second number is the smaller
        > equatorial radius, and the third is the polar radius.

        Example: Radii of the Earth.

        BODY399_RADII     = ( 6378.1366   6378.1366   6356.7519 )"""

    def flattening(self) -> float:
        """Returns the flattening ratio, computed from the mean equatorial radius and the polar radius"""

    def is_sphere(self) -> bool:
        """Returns true if the polar radius is equal to the semi minor radius."""

    def is_spheroid(self) -> bool:
        """Returns true if the semi major and minor radii are equal"""

    def mean_equatorial_radius_km(self) -> float:
        """Returns the mean equatorial radius in kilometers"""

    def __eq__(self, value: typing.Any) -> bool:
        """Return self==value."""

    def __ge__(self, value: typing.Any) -> bool:
        """Return self>=value."""

    def __getnewargs__(self) -> typing.Tuple:
        """Allows for pickling the object"""

    def __gt__(self, value: typing.Any) -> bool:
        """Return self>value."""

    def __le__(self, value: typing.Any) -> bool:
        """Return self<=value."""

    def __lt__(self, value: typing.Any) -> bool:
        """Return self<value."""

    def __ne__(self, value: typing.Any) -> bool:
        """Return self!=value."""

    def __repr__(self) -> str:
        """Return repr(self)."""

    def __str__(self) -> str:
        """Return str(self)."""

@typing.final
class Ephemeris:
    """Initializes a new Ephemeris from the list of Orbit instances and a given object ID.

    In Python if you need to build an ephemeris with covariance, initialize with an empty list of
    orbit instances and then insert each EphemerisRecord with covariance."""

    degree: int
    interpolation: str

    def __init__(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        """Initialize self.  See help(type(self)) for accurate signature."""

    def __new__(cls, orbit_list: list, object_id: str) -> Ephemeris:
        """Initializes a new Ephemeris from the list of Orbit instances and a given object ID.

        In Python if you need to build an ephemeris with covariance, initialize with an empty list of
        orbit instances and then insert each EphemerisRecord with covariance."""

    def at(self, epoch: time.Epoch, almanac: Almanac) -> EphemerisRecord:
        """Interpolates the ephemeris state and covariance at the provided epoch.

        # Orbit Interpolation
        The orbital state is interpolated using high-fidelity numeric methods consistent
        with SPICE standards:
        * **Type 9 (Lagrange):** Uses an Nth-order Lagrange polynomial interpolation on
        unequal time steps. It interpolates each of the 6 state components (position
        and velocity) independently.
        * **Type 13 (Hermite):** Uses an Nth-order Hermite interpolation. This method
        explicitly uses the velocity data (derivatives) to constrain the interpolation
        of the position, ensuring that the resulting position curve is smooth and
        dynamically consistent with the velocity.

        # Covariance Interpolation (Log-Euclidean)
        If covariance data is available, this method performs **Log-Euclidean Riemannian
        Interpolation**. Unlike standard linear element-wise interpolation, this approach
        respects the geometric manifold of Symmetric Positive Definite (SPD) matrices.

        This guarantees that:
        1. **Positive Definiteness:** The interpolated covariance matrix is always mathematically
        valid (all eigenvalues are strictly positive), preventing numerical crashes in downstream filters.
        2. **Volume Preservation:** It prevents the artificial "swelling" (determinant increase)
        of uncertainty that occurs when linearly interpolating between two valid matrices.
        The interpolation follows the "geodesic" (shortest path) on the curved surface of
        covariance matrices."""

    def covar_at(
        self, epoch: time.Epoch, local_frame: LocalFrame, almanac: Almanac
    ) -> Covariance:
        """Interpolate the ephemeris covariance at the provided epoch.

        This method implements a "Rotate-Then-Interpolate" strategy to avoid physical
        artifacts when interpolating rotating covariances.

        1. Finds the nearest covariance before and after the requested epoch.
        2. Rotates BOTH endpoints into the requested `local_frame`.
        3. Interpolates between the two stable matrices using Log-Euclidean Riemannian interpolation."""

    def domain(self) -> tuple:
        """Returns the time domain of this ephemeris."""

    def end_epoch(self) -> time.Epoch: ...
    @staticmethod
    def from_ccsds_oem_file(path: str) -> Ephemeris:
        """Initializes a new Ephemeris from a file path to CCSDS OEM file."""

    @staticmethod
    def from_stk_e_file(path: str) -> Ephemeris:
        """Initializes a new Ephemeris from a file path to Ansys STK .e file."""

    def includes_covariance(self) -> bool:
        """Returns true if all of the data in this ephemeris includes covariance.

        This is a helper function which isn't used in other functions."""

    def insert(self, record: EphemerisRecord) -> None:
        """Inserts a new ephemeris entry to this ephemeris (it is automatically sorted chronologically)."""

    def insert_orbit(self, orbit: Orbit) -> None:
        """Inserts a new orbit (without covariance) to this ephemeris (it is automatically sorted chronologically)."""

    def len(self) -> int:
        """Returns the number of states"""

    def nearest_after(self, epoch: time.Epoch, almanac: Almanac) -> EphemerisRecord:
        """Returns the nearest entry after the provided time"""

    def nearest_before(self, epoch: time.Epoch, almanac: Almanac) -> EphemerisRecord:
        """Returns the nearest entry before the provided time"""

    def nearest_covar_after(self, epoch: time.Epoch, almanac: Almanac) -> tuple:
        """Returns the nearest covariance after the provided epoch as a tuple (Epoch, Covariance)"""

    def nearest_covar_before(self, epoch: time.Epoch, almanac: Almanac) -> tuple:
        """Returns the nearest covariance before the provided epoch as a tuple (Epoch, Covariance)"""

    def nearest_orbit_after(self, epoch: time.Epoch, almanac: Almanac) -> Orbit:
        """Returns the nearest orbit after the provided time"""

    def nearest_orbit_before(self, epoch: time.Epoch, almanac: Almanac) -> Orbit:
        """Returns the nearest orbit before the provided time"""

    def object_id(self) -> str: ...
    def orbit_at(self, epoch: time.Epoch, almanac: Almanac) -> Orbit:
        """Interpolate the ephemeris at the provided epoch, returning only the orbit."""

    def resample(self, ts: time.TimeSeries, almanac: Almanac) -> Ephemeris:
        """Resample this ephemeris, with covariance, at the provided time series"""

    def start_epoch(self) -> time.Epoch: ...
    def transform(self, new_frame: Frame, almanac: Almanac) -> Ephemeris:
        """Transforms this ephemeris into another frame, and rotates the covariance to that frame if the orientations are different.
        NOTE: The Nyquist-Shannon theorem is NOT applied here, so the new ephemeris may not be as precise as the original one.
        NOTE: If the orientations are different, the covariance will always be in the Inertial frame of the new frame."""

    def write_ccsds_oem(
        self,
        path: str,
        originator: typing.Optional[str] = None,
        object_name: typing.Optional[str] = None,
    ) -> None:
        """Exports this Ephemeris to CCSDS OEM at the provided path, optionally specifying an originator and/or an object name"""

    def write_spice_bsp(
        self, naif_id: int, output_fname: str, data_type: DataType
    ) -> None:
        """Converts this ephemeris to SPICE BSP/SPK file in the provided data type, saved to the provided output_fname."""

    def __iter__(self) -> typing.Any:
        """Implement iter(self)."""

    def __repr__(self) -> str:
        """Return repr(self)."""

    def __reversed__(self): ...
    def __str__(self) -> str:
        """Return str(self)."""

@typing.final
class EphemerisRecord:
    """An ephemeris record which can be inserted into an Ephemeris for export to SPICE BSP or CCSDS OEM."""

    covar: Covariance
    orbit: Orbit

    def __init__(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        """Initialize self.  See help(type(self)) for accurate signature."""

    def __new__(cls, orbit: Orbit, covar: Covariance) -> EphemerisRecord:
        """An ephemeris record which can be inserted into an Ephemeris for export to SPICE BSP or CCSDS OEM."""

    def covar_in_frame(self, local_frame: LocalFrame) -> Covariance:
        """Returns the covariance in the desired orbit local frame, or None if this record does not define a covariance."""

    def sigma_for(self, oe: analysis.OrbitalElement) -> float:
        """Returns the 1-sigma uncertainty (Standard Deviation) for a given orbital element.

        The result is in the unit of the parameter (e.g., km for SMA, degrees for angles).

        This method uses the [OrbitGrad] structure (Hyperdual numbers) to compute the
        Jacobian of the element with respect to the inertial Cartesian state, and then
        rotates the covariance into that hyperdual dual space: J * P * J^T."""

@typing.final
class Frame:
    """A Frame uniquely defined by its ephemeris center and orientation. Refer to FrameDetail for frames combined with parameters."""

    ephemeris_id: int
    orientation_id: int
    shape: Ellipsoid

    def __init__(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        """Initialize self.  See help(type(self)) for accurate signature."""

    def __new__(
        cls,
        ephemeris_id: int,
        orientation_id: int,
        mu_km3_s2: typing.Optional[float] = None,
        shape: typing.Optional[Ellipsoid] = None,
    ) -> Frame:
        """A Frame uniquely defined by its ephemeris center and orientation. Refer to FrameDetail for frames combined with parameters."""

    def ephem_origin_id_match(self, other_id: int) -> bool:
        """Returns true if the ephemeris origin is equal to the provided ID"""

    def ephem_origin_match(self, other: Frame) -> bool:
        """Returns true if the ephemeris origin is equal to the provided frame"""

    def flattening(self) -> float:
        """Returns the flattening ratio (unitless)"""

    @staticmethod
    def from_asn1(data: bytes) -> Frame:
        """Decodes an ASN.1 DER encoded byte array into a Frame."""

    def is_celestial(self) -> bool:
        """Returns whether this is a celestial frame"""

    def is_geodetic(self) -> bool:
        """Returns whether this is a geodetic frame"""

    def mean_equatorial_radius_km(self) -> float:
        """Returns the mean equatorial radius in km, if defined"""

    def mu_km3_s2(self) -> float:
        """Returns the gravitational parameters of this frame, if defined"""

    def orient_origin_id_match(self, other_id: int) -> bool:
        """Returns true if the orientation origin is equal to the provided ID"""

    def orient_origin_match(self, other: Frame) -> bool:
        """Returns true if the orientation origin is equal to the provided frame"""

    def polar_radius_km(self) -> float:
        """Returns the polar radius in km, if defined"""

    def semi_major_radius_km(self) -> float:
        """Returns the semi major radius of the tri-axial ellipoid shape of this frame, if defined"""

    def strip(self) -> None:
        """Removes the graviational parameter and the shape information from this frame.
        Use this to prevent astrodynamical computations."""

    def to_asn1(self) -> bytes:
        """Encodes this Frame into an ASN.1 DER encoded byte array."""

    def with_ephem(self, new_ephem_id: int) -> Frame:
        """Returns a copy of this Frame whose ephemeris ID is set to the provided ID"""

    def with_mu_km3_s2(self, mu_km3_s2: float) -> Frame:
        """Returns a copy of this frame with the graviational parameter set to the new value."""

    def with_orient(self, new_orient_id: int) -> Frame:
        """Returns a copy of this Frame whose orientation ID is set to the provided ID"""

    def __eq__(self, value: typing.Any) -> bool:
        """Return self==value."""

    def __ge__(self, value: typing.Any) -> bool:
        """Return self>=value."""

    def __getnewargs__(self) -> typing.Tuple:
        """Allows for pickling the object"""

    def __gt__(self, value: typing.Any) -> bool:
        """Return self>value."""

    def __le__(self, value: typing.Any) -> bool:
        """Return self<=value."""

    def __lt__(self, value: typing.Any) -> bool:
        """Return self<value."""

    def __ne__(self, value: typing.Any) -> bool:
        """Return self!=value."""

    def __repr__(self) -> str:
        """Return repr(self)."""

    def __str__(self) -> str:
        """Return str(self)."""

@typing.final
class FrameUid:
    """A unique frame reference that only contains enough information to build the actual Frame object.
    It cannot be used for any computations, is it be used in any structure apart from error structures."""

    def __init__(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        """Initialize self.  See help(type(self)) for accurate signature."""

    def __new__(cls, ephemeris_id: int, orientation_id: int) -> FrameUid:
        """A unique frame reference that only contains enough information to build the actual Frame object.
        It cannot be used for any computations, is it be used in any structure apart from error structures."""

@typing.final
class LocalFrame:
    def __init__(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        """Initialize self.  See help(type(self)) for accurate signature."""

    def __new__(cls): ...
    def __int__(self) -> None:
        """int(self)"""

    def __repr__(self) -> str:
        """Return repr(self)."""
    Inertial: LocalFrame = ...
    RCN: LocalFrame = ...
    RIC: LocalFrame = ...
    VNC: LocalFrame = ...

@typing.final
class Location:
    """Location is defined by its latitude, longitude, height above the geoid, mean angular rotation of the geoid, and a frame UID.
    If the location includes a terrain mask, it will be used for obstruction checks when computing azimuth and elevation.
    **Note:** The mean Earth angular velocity is `0.004178079012116429` deg/s."""

    height_km: float
    latitude_deg: float
    longitude_deg: float
    terrain_mask: list
    terrain_mask_ignored: bool

    def __init__(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        """Initialize self.  See help(type(self)) for accurate signature."""

    def __new__(
        cls,
        latitude_deg: float,
        longitude_deg: float,
        height_km: float,
        frame: FrameUid,
        terrain_mask: list,
        terrain_mask_ignored: bool,
    ) -> Location:
        """Location is defined by its latitude, longitude, height above the geoid, mean angular rotation of the geoid, and a frame UID.
        If the location includes a terrain mask, it will be used for obstruction checks when computing azimuth and elevation.
        **Note:** The mean Earth angular velocity is `0.004178079012116429` deg/s."""

    def elevation_mask_at_azimuth_deg(self, azimuth_deg: float) -> float:
        """Returns the elevation mask at the provided azimuth, does NOT account for whether the mask is ignored or not."""

    @staticmethod
    def from_dhall(repr: str) -> Location:
        """Loads a Location from its Dhall representation"""

    def to_dhall(self) -> str:
        """Returns the Dhall representation of this Location"""

    def __repr__(self) -> str:
        """Return repr(self)."""

    def __str__(self) -> str:
        """Return str(self)."""

@typing.final
class Mass:
    """Defines a spacecraft mass a the sum of the dry (structural) mass and the propellant mass, both in kilogram"""

    dry_mass_kg: float
    extra_mass_kg: float
    prop_mass_kg: float

    def __init__(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        """Initialize self.  See help(type(self)) for accurate signature."""

    def __new__(
        cls,
        dry_mass_kg: typing.Any,
        prop_mass_kg: typing.Any = None,
        extra_mass_kg: typing.Any = None,
    ) -> Mass:
        """Defines a spacecraft mass a the sum of the dry (structural) mass and the propellant mass, both in kilogram"""

    def abs(self) -> Mass:
        """Returns a Mass structure that is guaranteed to be physically correct"""

    @staticmethod
    def from_asn1(data: bytes) -> Mass:
        """Decodes an ASN.1 DER encoded byte array into a Mass object."""

    def is_valid(self) -> bool:
        """Returns true if all the masses are greater or equal to zero"""

    def to_asn1(self) -> bytes:
        """Encodes this Mass object into an ASN.1 DER encoded byte array."""

    def total_mass_kg(self) -> float:
        """Returns the total mass in kg"""

    def __repr__(self) -> str:
        """Return repr(self)."""

    def __str__(self) -> str:
        """Return str(self)."""

@typing.final
class Occultation:
    """Stores the result of an occultation computation with the occultation percentage
    Refer to the [MathSpec](https://nyxspace.com/nyxspace/MathSpec/celestial/eclipse/) for modeling details."""

    back_frame: Frame
    epoch: time.Epoch
    front_frame: Frame
    percentage: float

    def __init__(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        """Initialize self.  See help(type(self)) for accurate signature."""

    def __new__(cls) -> Occultation:
        """Stores the result of an occultation computation with the occultation percentage
        Refer to the [MathSpec](https://nyxspace.com/nyxspace/MathSpec/celestial/eclipse/) for modeling details."""

    def factor(self) -> float:
        """Returns the percentage as a factor between 0 and 1"""

    def is_eclipse_computation(self) -> bool:
        """Returns true if the back object is the Sun, false otherwise"""

    def is_obstructed(self) -> bool:
        """Returns true if the occultation percentage is greater than or equal 99.999%"""

    def is_partial(self) -> bool:
        """Returns true if neither occulted nor visible (i.e. penumbra for solar eclipsing)"""

    def is_visible(self) -> bool:
        """Returns true if the occultation percentage is less than or equal 0.001%"""

    def __repr__(self) -> str:
        """Return repr(self)."""

    def __str__(self) -> str:
        """Return str(self)."""

@typing.final
class Orbit:
    """Defines a Cartesian state in a given frame at a given epoch in a given time scale. Radius data is expressed in kilometers. Velocity data is expressed in kilometers per second.
    Regardless of the constructor used, this struct stores all the state information in Cartesian coordinates as these are always non singular.

    Unless noted otherwise, algorithms are from GMAT 2016a [StateConversionUtil.cpp](https://github.com/ChristopherRabotin/GMAT/blob/37201a6290e7f7b941bc98ee973a527a5857104b/src/base/util/StateConversionUtil.cpp)."""

    epoch: time.Epoch
    frame: Frame
    vx_km_s: float
    vy_km_s: float
    vz_km_s: float
    x_km: float
    y_km: float
    z_km: float

    def __init__(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        """Initialize self.  See help(type(self)) for accurate signature."""

    def __new__(cls, *args: typing.Tuple) -> Orbit:
        """Defines a Cartesian state in a given frame at a given epoch in a given time scale. Radius data is expressed in kilometers. Velocity data is expressed in kilometers per second.
        Regardless of the constructor used, this struct stores all the state information in Cartesian coordinates as these are always non singular.

        Unless noted otherwise, algorithms are from GMAT 2016a [StateConversionUtil.cpp](https://github.com/ChristopherRabotin/GMAT/blob/37201a6290e7f7b941bc98ee973a527a5857104b/src/base/util/StateConversionUtil.cpp)."""

    def abs_difference(self, other: Orbit) -> typing.Tuple:
        """Returns the absolute position and velocity differences in km and km/s between this orbit and another.
        Raises an error if the frames do not match (epochs do not need to match)."""

    def abs_pos_diff_km(self, other: Orbit) -> float:
        """Returns the absolute position difference in kilometer between this orbit and another.
        Raises an error if the frames do not match (epochs do not need to match)."""

    def abs_vel_diff_km_s(self, other: Orbit) -> float:
        """Returns the absolute velocity difference in kilometer per second between this orbit and another.
        Raises an error if the frames do not match (epochs do not need to match)."""

    def add_aop_deg(self, delta_aop_deg: float) -> Orbit:
        """Returns a copy of the state with a provided AOP added to the current one"""

    def add_apoapsis_periapsis_km(
        self, delta_ra_km: float, delta_rp_km: float
    ) -> Orbit:
        """Returns a copy of this state with the provided apoasis and periapsis added to the current values"""

    def add_ecc(self, delta_ecc: float) -> Orbit:
        """Returns a copy of the state with a provided ECC added to the current one"""

    def add_inc_deg(self, delta_inc_deg: float) -> Orbit:
        """Returns a copy of the state with a provided INC added to the current one"""

    def add_raan_deg(self, delta_raan_deg: float) -> Orbit:
        """Returns a copy of the state with a provided RAAN added to the current one"""

    def add_sma_km(self, delta_sma_km: float) -> Orbit:
        """Returns a copy of the state with a provided SMA added to the current one"""

    def add_ta_deg(self, delta_ta_deg: float) -> Orbit:
        """Returns a copy of the state with a provided TA added to the current one"""

    def altitude_km(self) -> float:
        """Returns the altitude in km"""

    def aol_deg(self) -> float:
        """Returns the argument of latitude in degrees

        NOTE: If the orbit is near circular, the AoL will be computed from the true longitude
        instead of relying on the ill-defined true anomaly."""

    def aop_brouwer_short_deg(self) -> float:
        """Returns the Brouwer-short mean Argument of Perigee in degrees."""

    def aop_deg(self) -> float:
        """Returns the argument of periapsis in degrees"""

    def apoapsis_altitude_km(self) -> float:
        """Returns the altitude of apoapsis (or apogee around Earth), in kilometers."""

    def apoapsis_km(self) -> float:
        """Returns the radius of apoapsis (or apogee around Earth), in kilometers."""

    def at_epoch(self, new_epoch: time.Epoch) -> Orbit:
        """Adjusts the true anomaly of this orbit using the mean anomaly.

        # Astrodynamics note
        This is not a true propagation of the orbit. This is akin to a two body propagation ONLY without any other force models applied.
        Use Nyx for high fidelity propagation."""

    def c3_km2_s2(self) -> float:
        """Returns the $C_3$ of this orbit in km^2/s^2"""

    def cartesian_pos_vel(self) -> numpy.ndarray:
        """Returns this state as a Cartesian vector of size 6 in [km, km, km, km/s, km/s, km/s]

        Note that the time is **not** returned in the vector."""

    def dcm3x3_from_rcn_to_inertial(self) -> rotation.DCM:
        """Builds the rotation matrix that rotates from this state's inertial frame to this state's RCN frame (radial, cross, normal)

        # Frame warning
        If the stattion is NOT in an inertial frame, then this computation is INVALID.

        # Algorithm
        1. Compute \\hat{r}, \\hat{h}, the unit vectors of the radius and orbital momentum.
        2. Compute the cross product of these
        3. Build the DCM with these unit vectors
        4. Return the DCM structure"""

    def dcm3x3_from_ric_to_inertial(self) -> rotation.DCM:
        """Builds the rotation matrix that rotates from this state's inertial frame to this state's RIC frame

        # Frame warning
        If the state is NOT in an inertial frame, then this computation is INVALID.

        # Algorithm
        1. Build the c vector as the normalized orbital momentum vector
        2. Build the i vector as the cross product of \\hat{r} and c
        3. Build the RIC DCM as a 3x3 of the columns [\\hat{r}, \\hat{i}, \\hat{c}]
        4. Return the DCM structure **without** accounting for the transport theorem."""

    def dcm3x3_from_topocentric_to_body_fixed(self) -> rotation.DCM:
        """Builds the rotation matrix that rotates from the topocentric frame (SEZ) into the body fixed frame of this state.

        # Frame warning
        If the state is NOT in a body fixed frame (i.e. ITRF93), then this computation is INVALID.

        # Source
        From the GMAT MathSpec, page 30 section 2.6.9 and from `Calculate_RFT` in `TopocentricAxes.cpp`, this returns the
        rotation matrix from the topocentric frame (SEZ) to body fixed frame.
        In the GMAT MathSpec notation, R_{IF} is the DCM from body fixed to inertial. Similarly, R{FT} is from topocentric
        to body fixed."""

    def dcm3x3_from_vnc_to_inertial(self) -> rotation.DCM:
        """Builds the rotation matrix that rotates from this state's inertial frame to this state's VNC frame (velocity, normal, cross)

        # Frame warning
        If the stattion is NOT in an inertial frame, then this computation is INVALID.

        # Algorithm
        1. Compute \\hat{v}, \\hat{h}, the unit vectors of the radius and orbital momentum.
        2. Compute the cross product of these
        3. Build the DCM with these unit vectors
        4. Return the DCM structure."""

    def dcm_from_rcn_to_inertial(self) -> rotation.DCM:
        """Builds the rotation matrix that rotates from this state's inertial frame to this state's RCN frame (radial, cross, normal)

        # Frame warning
        If the stattion is NOT in an inertial frame, then this computation is INVALID.

        # Algorithm
        1. Compute \\hat{r}, \\hat{h}, the unit vectors of the radius and orbital momentum.
        2. Compute the cross product of these
        3. Build the DCM with these unit vectors
        4. Return the DCM structure with a 6x6 DCM with the time derivative of the VNC frame set.

        # Note on the time derivative
        If the pre or post states cannot be computed, then the time derivative of the DCM will _not_ be set.
        Further note that most astrodynamics tools do *not* account for the time derivative in the RIC frame."""

    def dcm_from_ric_to_inertial(self) -> rotation.DCM:
        """Builds the rotation matrix that rotates from this state's inertial frame to this state's RIC frame

        # Frame warning
        If the state is NOT in an inertial frame, then this computation is INVALID.

        # Algorithm
        1. Compute the state data one millisecond before and one millisecond assuming two body dynamics
        2. Compute the DCM for this state, and the pre and post states
        3. Build the c vector as the normalized orbital momentum vector
        4. Build the i vector as the cross product of \\hat{r} and c
        5. Build the RIC DCM as a 3x3 of the columns [\\hat{r}, \\hat{i}, \\hat{c}], for the post, post, and current states
        6. Compute the difference between the DCMs of the pre and post states, to build the DCM time derivative
        7. Return the DCM structure with a 6x6 state DCM.

        # Note on the time derivative
        If the pre or post states cannot be computed, then the time derivative of the DCM will _not_ be set.
        Further note that most astrodynamics tools do *not* account for the time derivative in the RIC frame."""

    def dcm_from_topocentric_to_body_fixed(self) -> rotation.DCM:
        """Builds the rotation matrix that rotates from the topocentric frame (SEZ) into the body fixed frame of this state.

        # Frame warnings
        + If the state is NOT in a body fixed frame (i.e. ITRF93), then this computation is INVALID.
        + (Usually) no time derivative can be computed: the orbit is expected to be a body fixed frame where the `at_epoch` function will fail. Exceptions for Moon body fixed frames.

        # UNUSED Arguments
        + `from`: ID of this new frame. Only used to set the "from" frame of the DCM. -- No longer used since 0.5.3

        # Source
        From the GMAT MathSpec, page 30 section 2.6.9 and from `Calculate_RFT` in `TopocentricAxes.cpp`, this returns the
        rotation matrix from the topocentric frame (SEZ) to body fixed frame.
        In the GMAT MathSpec notation, R_{IF} is the DCM from body fixed to inertial. Similarly, R{FT} is from topocentric
        to body fixed."""

    def dcm_from_vnc_to_inertial(self) -> rotation.DCM:
        """Builds the rotation matrix that rotates from this state's inertial frame to this state's VNC frame (velocity, normal, cross)

        # Frame warning
        If the stattion is NOT in an inertial frame, then this computation is INVALID.

        # Algorithm
        1. Compute \\hat{v}, \\hat{h}, the unit vectors of the radius and orbital momentum.
        2. Compute the cross product of these
        3. Build the DCM with these unit vectors
        4. Compute the difference between the DCMs of the pre and post states (+/- 1 ms), to build the DCM time derivative
        4. Return the DCM structure with a 6x6 DCM with the time derivative of the VNC frame set.

        # Note on the time derivative
        If the pre or post states cannot be computed, then the time derivative of the DCM will _not_ be set.
        Further note that most astrodynamics tools do *not* account for the time derivative in the RIC frame."""

    def dcm_to_inertial(self, local_frame: LocalFrame) -> rotation.DCM:
        """Returns the DCM to rotate this orbit from the provided local frame to the inertial frame."""

    def declination_deg(self) -> float:
        """Returns the declination of this orbit in degrees"""

    def distance_to_km(self, other: Orbit) -> float:
        """Returns the distance in kilometers between this state and another state, if both frame match (epoch does not need to match)."""

    def duration_to_radius(self, radius_km: float) -> time.Duration:
        """Calculates the duration to reach a specific radius in the orbit.

        This function computes the time it will take for the orbiting body to reach
        the given `radius_km` from its current position. The calculation assumes
        two-body dynamics and considers the direction of motion.

        # Assumptions & Limitations

        - Assumes pure Keplerian motion.
        - For elliptical orbits, if the radius is reachable at two points (ascending and descending parts
        of the orbit), this function calculates the time to reach the radius corresponding to the
        true anomaly in `[0, PI]` (typically the ascending part or up to apoapsis if starting past periapsis).
        - For circular orbits, if the radius is within the apoapse and periapse, then a duration of zero is returned.
        - For hyperbolic/parabolic orbits, the true anomaly at radius is also computed in `[0, PI]`. If this
        point is in the past, the function returns an error, as it doesn't look for solutions on the
        departing leg if `nu > PI` would be required (unless current TA is already > PI and target radius is further along).
        The current implementation strictly uses the `acos` result, so `nu_rad_at_radius` is always `0 <= nu <= PI`.
        This means it finds the time to reach the radius on the path from periapsis up to the point where true anomaly is PI."""

    def ea_deg(self) -> float:
        """Returns the eccentric anomaly in degrees

        This is a conversion from GMAT's StateConversionUtil::TrueToEccentricAnomaly"""

    def ecc(self) -> float:
        """Returns the eccentricity (no unit)"""

    def ecc_brouwer_short(self) -> float:
        """Returns the Brouwer-short mean eccentricity."""

    def energy_km2_s2(self) -> float:
        """Returns the specific mechanical energy in km^2/s^2"""

    def eq_within(
        self, other: Orbit, radial_tol_km: float, velocity_tol_km_s: float
    ) -> bool:
        """Returns whether this orbit and another are equal within the specified radial and velocity absolute tolerances"""

    def equinoctial_a_km(self) -> float:
        """Returns the equinoctial semi-major axis (a) in km."""

    def equinoctial_h(self) -> float:
        """Returns the equinoctial element h (ecc * sin(aop + raan))."""

    def equinoctial_k(self) -> float:
        """Returns the equinoctial element k (ecc * cos(aop + raan))."""

    def equinoctial_lambda_mean_deg(self) -> float:
        """Returns the equinoctial mean longitude (lambda = raan + aop + ma) in degrees."""

    def equinoctial_p(self) -> float:
        """Returns the equinoctial element p (sin(inc/2) * sin(raan))."""

    def equinoctial_q(self) -> float:
        """Returns the equinoctial element q (sin(inc/2) * cos(raan))."""

    def fpa_deg(self) -> float:
        """Returns the flight path angle in degrees"""

    @staticmethod
    def from_asn1(data: bytes) -> Orbit:
        """Decodes an ASN.1 DER encoded byte array into a CartesianState (Orbit)."""

    @staticmethod
    def from_cartesian(
        x_km: float,
        y_km: float,
        z_km: float,
        vx_km_s: float,
        vy_km_s: float,
        vz_km_s: float,
        epoch: time.Epoch,
        frame: Frame,
    ) -> Orbit:
        """Creates a new Cartesian state in the provided frame at the provided Epoch.

        **Units:** km, km, km, km/s, km/s, km/s"""

    @staticmethod
    def from_cartesian_npy(
        pos_vel: numpy.ndarray, epoch: time.Epoch, frame: Frame
    ) -> Orbit:
        """Creates a new Cartesian state from a numpy array, an epoch, and a frame.

        **Units:** km, km, km, km/s, km/s, km/s"""

    @staticmethod
    def from_keplerian(
        sma_km: float,
        ecc: float,
        inc_deg: float,
        raan_deg: float,
        aop_deg: float,
        ta_deg: float,
        epoch: time.Epoch,
        frame: Frame,
    ) -> Orbit:
        """Creates a new Orbit around the provided Celestial or Geoid frame from the Keplerian orbital elements.

        **Units:** km, none, degrees, degrees, degrees, degrees

        NOTE: The state is defined in Cartesian coordinates as they are non-singular. This causes rounding
        errors when creating a state from its Keplerian orbital elements (cf. the state tests).
        One should expect these errors to be on the order of 1e-12."""

    @staticmethod
    def from_keplerian_altitude(
        sma_altitude_km: float,
        ecc: float,
        inc_deg: float,
        raan_deg: float,
        aop_deg: float,
        ta_deg: float,
        epoch: time.Epoch,
        frame: Frame,
    ) -> Orbit:
        """Creates a new Orbit from the provided semi-major axis altitude in kilometers"""

    @staticmethod
    def from_keplerian_apsis_altitude(
        apo_alt_km: float,
        peri_alt_km: float,
        inc_deg: float,
        raan_deg: float,
        aop_deg: float,
        ta_deg: float,
        epoch: time.Epoch,
        frame: Frame,
    ) -> Orbit:
        """Creates a new Orbit from the provided altitudes of apoapsis and periapsis, in kilometers"""

    @staticmethod
    def from_keplerian_apsis_radii(
        r_a_km: float,
        r_p_km: float,
        inc_deg: float,
        raan_deg: float,
        aop_deg: float,
        ta_deg: float,
        epoch: time.Epoch,
        frame: Frame,
    ) -> Orbit:
        """Attempts to create a new Orbit from the provided radii of apoapsis and periapsis, in kilometers"""

    @staticmethod
    def from_keplerian_mean_anomaly(
        sma_km: float,
        ecc: float,
        inc_deg: float,
        raan_deg: float,
        aop_deg: float,
        ma_deg: float,
        epoch: time.Epoch,
        frame: Frame,
    ) -> Orbit:
        """Initializes a new orbit from the Keplerian orbital elements using the mean anomaly instead of the true anomaly.

        # Implementation notes
        This function starts by converting the mean anomaly to true anomaly, and then it initializes the orbit
        using the keplerian(..) method.
        The conversion is from GMAT's MeanToTrueAnomaly function, transliterated originally by Claude and GPT4 with human adjustments."""

    @staticmethod
    def from_latlongalt(
        latitude_deg: float,
        longitude_deg: float,
        height_km: float,
        epoch: time.Epoch,
        frame: Frame,
    ) -> Orbit:
        """Creates a new Orbit from the latitude (φ), longitude (λ) and height (in km) with respect to the frame's ellipsoid, and with ZERO angular velocity in this frame.
        Use this initializer for creating a fixed point in the ITRF93 frame for example: the correct angular velocity will be applied when transforming this to EME2000 for example.

        Refer to [try_latlongalt_omega] if you need to build a fixed point with a non-zero angular velocity in the definition frame.

        NOTE: This computation differs from the spherical coordinates because we consider the flattening of body.
        Reference: G. Xu and Y. Xu, "GPS", DOI 10.1007/978-3-662-50367-6_2, 2016"""

    @staticmethod
    def from_latlongalt_omega(
        latitude_deg: float,
        longitude_deg: float,
        height_km: float,
        angular_velocity_rad_s: numpy.ndarray,
        epoch: time.Epoch,
        frame: Frame,
    ) -> Orbit:
        """Creates a new Orbit from the latitude (φ), longitude (λ) and height (in km) with respect to the frame's ellipsoid given the angular velocity vector.
        NOTE: Only specify the angular velocity if there's an EXTRA angular velocity of the lat/long/alt point in the provided frame.

        Consider using the [Almanac]'s [angular_velocity_wrt_j2000_rad_s] function or [angular_velocity_rad_s] to retrieve the exact angular velocity vector between two orientations.
        Example: build a lat/long/alt point referenced in the ITRF93 frame but by specifying the Frame as the EME2000 frame and providing the angular velocity between the ITRF93 and EME2000 frame at the desired time.

        NOTE: This computation differs from the spherical coordinates because we consider the flattening of body.
        Reference: G. Xu and Y. Xu, "GPS", DOI 10.1007/978-3-662-50367-6_2, 2016"""

    def height_km(self) -> float:
        """Returns the geodetic height in km.

        Reference: Vallado, 4th Ed., Algorithm 12 page 172."""

    def hmag(self) -> float:
        """Returns the norm of the orbital momentum"""

    def hx(self) -> float:
        """Returns the orbital momentum value on the X axis"""

    def hy(self) -> float:
        """Returns the orbital momentum value on the Y axis"""

    def hyperbolic_anomaly_deg(self) -> float:
        """Returns the hyperbolic anomaly in degrees between 0 and 360.0
        Returns an error if the orbit is not hyperbolic."""

    def hz(self) -> float:
        """Returns the orbital momentum value on the Z axis"""

    def inc_brouwer_short_deg(self) -> float:
        """Returns the Brouwer-short mean inclination in degrees."""

    def inc_deg(self) -> float:
        """Returns the inclination in degrees"""

    def is_brouwer_short_valid(self) -> bool:
        """Returns whether this state satisfies the requirement to compute the Mean Brouwer Short orbital
        element set.

        This is a conversion from GMAT's StateConversionUtil::CartesianToBrouwerMeanShort.
        The details are at the log level `info`.
        NOTE: Mean Brouwer Short are only defined around Earth. However, `nyx` does *not* check the
        main celestial body around which the state is defined (GMAT does perform this verification)."""

    def latitude_deg(self) -> float:
        """Returns the geodetic latitude (φ) in degrees. Value is between -180 and +180 degrees.

        # Frame warning
        This state MUST be in the body fixed frame (e.g. ITRF93) prior to calling this function, or the computation is **invalid**."""

    def latlongalt(self) -> typing.Tuple:
        """Returns the geodetic latitude, geodetic longitude, and geodetic height, respectively in degrees, degrees, and kilometers.

        # Algorithm
        This uses the Heikkinen procedure, which is not iterative. The results match Vallado and GMAT."""

    def light_time(self) -> time.Duration:
        """Returns the light time duration between this object and the origin of its reference frame."""

    def longitude_360_deg(self) -> float:
        """Returns the geodetic longitude (λ) in degrees. Value is between 0 and 360 degrees.

        # Frame warning
        This state MUST be in the body fixed frame (e.g. ITRF93) prior to calling this function, or the computation is **invalid**."""

    def longitude_deg(self) -> float:
        """Returns the geodetic longitude (λ) in degrees. Value is between -180 and 180 degrees.

        # Frame warning
        This state MUST be in the body fixed frame (e.g. ITRF93) prior to calling this function, or the computation is **invalid**."""

    def ltan_deg(self) -> float:
        """Returns the Longitude of the Ascending Node (LTAN), or an error of equatorial orbits"""

    def ma_brouwer_short_deg(self) -> float:
        """Returns the Brouwer-short mean Mean Anomaly in degrees."""

    def ma_deg(self) -> float:
        """Returns the mean anomaly in degrees

        This is a conversion from GMAT's StateConversionUtil::TrueToMeanAnomaly"""

    def mean_motion_deg_s(self) -> float:
        """Returns the mean motion in degrees per seconds"""

    def periapsis_altitude_km(self) -> float:
        """Returns the altitude of periapsis (or perigee around Earth), in kilometers."""

    def periapsis_km(self) -> float:
        """Returns the radius of periapsis (or perigee around Earth), in kilometers."""

    def period(self) -> time.Duration:
        """Returns the period"""

    def raan_brouwer_short_deg(self) -> float:
        """Returns the Brouwer-short mean Right Ascension of the Ascending Node in degrees."""

    def raan_deg(self) -> float:
        """Returns the right ascension of the ascending node in degrees"""

    def radius_km(self) -> numpy.ndarray:
        """radius vector in km"""

    def rel_difference(self, other: Orbit) -> typing.Tuple:
        """Returns the relative difference between this orbit and another for the position and velocity, respectively the first and second return values.
        Both return values are UNITLESS because the relative difference is computed as the absolute difference divided by the rmag and vmag of this object.
        Raises an error if the frames do not match, if the position is zero or the velocity is zero."""

    def rel_pos_diff(self, other: Orbit) -> float:
        """Returns the relative position difference (unitless) between this orbit and another.
        This is computed by dividing the absolute difference by the norm of this object's radius vector.
        If the radius is zero, this function raises a math error.
        Raises an error if the frames do not match or  (epochs do not need to match)."""

    def rel_vel_diff(self, other: Orbit) -> float:
        """Returns the absolute velocity difference in kilometer per second between this orbit and another.
        Raises an error if the frames do not match (epochs do not need to match)."""

    def ric_difference(self, other: Orbit) -> Orbit:
        """Returns a Cartesian state representing the RIC difference between self and other, in position and velocity (with transport theorem).
        Refer to dcm_from_ric_to_inertial for details on the RIC frame.

        # Algorithm
        1. Compute the difference between `other` and `self`
        2. Compute the RIC DCM of `self`
        3. Rotate the difference into the RIC frame of `self`
        4. Strip the astrodynamical information from the frame, enabling only computations from `CartesianState`"""

    def right_ascension_deg(self) -> float:
        """Returns the right ascension of this orbit in degrees"""

    def rmag_km(self) -> float:
        """Returns the magnitude of the radius vector in km"""

    def rms_radius_km(self, other: Orbit) -> float:
        """Returns the root sum squared (RMS) radius difference between this state and another state, if both frames match (epoch does not need to match)"""

    def rms_velocity_km_s(self, other: Orbit) -> float:
        """Returns the root sum squared (RMS) velocity difference between this state and another state, if both frames match (epoch does not need to match)"""

    def rss_radius_km(self, other: Orbit) -> float:
        """Returns the root mean squared (RSS) radius difference between this state and another state, if both frames match (epoch does not need to match)"""

    def rss_velocity_km_s(self, other: Orbit) -> float:
        """Returns the root mean squared (RSS) velocity difference between this state and another state, if both frames match (epoch does not need to match)"""

    def semi_minor_axis_km(self) -> float:
        """Returns the semi minor axis in km, includes code for a hyperbolic orbit"""

    def semi_parameter_km(self) -> float:
        """Returns the semi parameter (or semilatus rectum)"""

    def set_aop_deg(self, new_aop_deg: float) -> None:
        """Mutates this orbit to change the AOP"""

    def set_ecc(self, new_ecc: float) -> None:
        """Mutates this orbit to change the ECC"""

    def set_inc_deg(self, new_inc_deg: float) -> None:
        """Mutates this orbit to change the INC"""

    def set_raan_deg(self, new_raan_deg: float) -> None:
        """Mutates this orbit to change the RAAN"""

    def set_sma_km(self, new_sma_km: float) -> None:
        """Mutates this orbit to change the SMA"""

    def set_ta_deg(self, new_ta_deg: float) -> None:
        """Mutates this orbit to change the TA"""

    def sma_altitude_km(self) -> float:
        """Returns the SMA altitude in km"""

    def sma_brouwer_short_km(self) -> float:
        """Returns the Brouwer-short mean semi-major axis in km."""

    def sma_km(self) -> float:
        """Returns the semi-major axis in km"""

    def ta_deg(self) -> float:
        """Returns the true anomaly in degrees between 0 and 360.0

        NOTE: This function will emit a warning stating that the TA should be avoided if in a very near circular orbit
        Code from <https://github.com/ChristopherRabotin/GMAT/blob/80bde040e12946a61dae90d9fc3538f16df34190/src/gmatutil/util/StateConversionUtil.cpp#L6835>

        LIMITATION: For an orbit whose true anomaly is (very nearly) 0.0 or 180.0, this function may return either 0.0 or 180.0 with a very small time increment.
        This is due to the precision of the cosine calculation: if the arccosine calculation is out of bounds, the sign of the cosine of the true anomaly is used
        to determine whether the true anomaly should be 0.0 or 180.0. **In other words**, there is an ambiguity in the computation in the true anomaly exactly at 180.0 and 0.0."""

    def ta_dot_deg_s(self) -> float:
        """Returns the time derivative of the true anomaly computed as the 360.0 degrees divided by the orbital period (in seconds)."""

    def tlong_deg(self) -> float:
        """Returns the true longitude in degrees"""

    def to_asn1(self) -> bytes:
        """Encodes this CartesianState (Orbit) into an ASN.1 DER encoded byte array."""

    def velocity_declination_deg(self) -> float:
        """Returns the velocity declination of this orbit in degrees"""

    def velocity_km_s(self) -> numpy.ndarray:
        """velocity vector in km/s"""

    def vinf_periapsis_km(self, turn_angle_degrees: float) -> float:
        """Returns the radius of periapse in kilometers for the provided turn angle of this hyperbolic orbit.
        Returns an error if the orbit is not hyperbolic."""

    def vinf_turn_angle_deg(self, periapsis_km: float) -> float:
        """Returns the turn angle in degrees for the provided radius of periapse passage of this hyperbolic orbit
        Returns an error if the orbit is not hyperbolic."""

    def vmag_km_s(self) -> float:
        """Returns the magnitude of the velocity vector in km/s"""

    def vnc_difference(self, other: Orbit) -> Orbit:
        """Returns a Cartesian state representing the VNC difference between self and other, in position and velocity (with transport theorem).
        Refer to dcm_from_vnc_to_inertial for details on the VNC frame.

        # Algorithm
        1. Compute the difference between `other` and `self`
        2. Compute the VNC DCM of `self`
        3. Rotate the difference into the VNC frame of `self`
        4. Strip the astrodynamical information from the frame, enabling only computations from `CartesianState`"""

    def with_aop_deg(self, new_aop_deg: float) -> Orbit:
        """Returns a copy of the state with a new AOP"""

    def with_apoapsis_periapsis_km(self, new_ra_km: float, new_rp_km: float) -> Orbit:
        """Returns a copy of this state with the provided apoasis and periapsis"""

    def with_ecc(self, new_ecc: float) -> Orbit:
        """Returns a copy of the state with a new ECC"""

    def with_inc_deg(self, new_inc_deg: float) -> Orbit:
        """Returns a copy of the state with a new INC"""

    def with_raan_deg(self, new_raan_deg: float) -> Orbit:
        """Returns a copy of the state with a new RAAN"""

    def with_sma_km(self, new_sma_km: float) -> Orbit:
        """Returns a copy of the state with a new SMA"""

    def with_ta_deg(self, new_ta_deg: float) -> Orbit:
        """Returns a copy of the state with a new TA"""

    def __eq__(self, value: typing.Any) -> bool:
        """Return self==value."""

    def __ge__(self, value: typing.Any) -> bool:
        """Return self>=value."""

    def __getnewargs__(self) -> typing.Tuple: ...
    def __gt__(self, value: typing.Any) -> bool:
        """Return self>value."""

    def __le__(self, value: typing.Any) -> bool:
        """Return self<=value."""

    def __lt__(self, value: typing.Any) -> bool:
        """Return self<value."""

    def __ne__(self, value: typing.Any) -> bool:
        """Return self!=value."""

    def __repr__(self) -> str:
        """Return repr(self)."""

    def __str__(self) -> str:
        """Return str(self)."""

@typing.final
class SRPData:
    area_m2: float
    coeff_reflectivity: float

    def __init__(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        """Initialize self.  See help(type(self)) for accurate signature."""

    def __new__(cls, area_m2: typing.Any, coeff_reflectivity: typing.Any = None): ...
    @staticmethod
    def from_asn1(data: bytes) -> SRPData:
        """Decodes an ASN.1 DER encoded byte array into an SRPData object."""

    def to_asn1(self) -> bytes:
        """Encodes this SRPData object into an ASN.1 DER encoded byte array."""

    def __repr__(self) -> str:
        """Return repr(self)."""

    def __str__(self) -> str:
        """Return str(self)."""

@typing.final
class TerrainMask:
    """TerrainMask is used to compute obstructions during AER calculations."""

    azimuth_deg: float
    elevation_mask_deg: float

    def __init__(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        """Initialize self.  See help(type(self)) for accurate signature."""

    def __new__(cls, azimuth_deg: float, elevation_mask_deg: float) -> TerrainMask:
        """TerrainMask is used to compute obstructions during AER calculations."""

    @staticmethod
    def from_flat_terrain(elevation_mask_deg: float) -> list:
        """Creates a flat terrain mask with the provided elevation mask in degrees"""

    def __repr__(self) -> str:
        """Return repr(self)."""

    def __str__(self) -> str:
        """Return str(self)."""



import os
import pickle
from math import radians
from os import environ
from pathlib import Path

import numpy as np

from anise import (
    Almanac,
    LocationDataSet,
    LocationDhallSet,
    LocationDhallSetEntry,
    MetaAlmanac,
    MetaFile,
)
from anise.analysis import OrbitalElement
from anise.astro import (
    Covariance,
    DataType,
    Ellipsoid,
    Ephemeris,
    EphemerisRecord,
    Frame,
    FrameUid,
    LocalFrame,
    Location,
    Orbit,
    TerrainMask,
)
from anise.constants import Frames
from anise.rotation import DCM, Quaternion
from anise.time import Duration, Epoch, TimeSeries, Unit
from anise.utils import convert_tpc


def test_state_transformation():
    """
    This is the Python equivalent to anise/tests/almanac/mod.rs
    but the data is loaded from the remote servers
    """

    if environ.get("CI", False):
        # Load from meta kernel to not use Git LFS quota
        data_path = Path(__file__).parent.joinpath(
            "..", "..", "data", "ci_config.dhall"
        )
        meta = MetaAlmanac(str(data_path))
        print(meta)
        # Process the files to be loaded
        try:
            almanac = meta.process()
        except Exception as e:
            if "lfs" in str(e):
                # Must be some LFS error in the CI again
                return
            raise  # Otherwise, raise the error!
    else:
        data_path = Path(__file__).parent.joinpath("..", "..", "data")
        # Must ensure that the path is a string
        almanac = Almanac(str(data_path.joinpath("de440s.bsp")))
        # Let's add another file here -- note that the Almanac will load into a NEW variable, so we must overwrite it!
        # This prevents memory leaks (yes, I promise)
        almanac = almanac.load(str(data_path.joinpath("pck08.pca"))).load(
            str(data_path.joinpath("earth_latest_high_prec.bpc"))
        )

    eme2k = almanac.frame_info(Frames.EME2000)
    assert eme2k.mu_km3_s2() == 398600.435436096
    assert eme2k.shape.polar_radius_km == 6356.75
    assert abs(eme2k.shape.flattening() - 0.0033536422844278) < 2e-16

    epoch = Epoch("2021-10-29 12:34:56 TDB")

    orig_state = Orbit.from_keplerian(
        8_191.93,
        1e-6,
        12.85,
        306.614,
        314.19,
        99.887_7,
        epoch,
        eme2k,
    )

    assert abs(orig_state.sma_km() - 8191.93) < 1e-10
    assert abs(orig_state.ecc() - 1.000000000361619e-06) < 1e-10
    assert abs(orig_state.inc_deg() - 12.849999999999987) < 1e-10
    assert abs(orig_state.raan_deg() - 306.614) < 1e-10
    assert abs(orig_state.tlong_deg() - 0.6916999999999689) < 1e-10

    assert orig_state.cartesian_pos_vel().shape == (6,)

    # Check that we can serialize in ASN.1
    as_asn1 = orig_state.to_asn1()
    assert isinstance(as_asn1, bytes)
    assert Orbit.from_asn1(as_asn1) == orig_state

    # Ensure we can call all of the DCM functions
    for func in [
        "dcm_from_ric_to_inertial",
        "dcm_from_rcn_to_inertial",
        "dcm_from_vnc_to_inertial",
    ]:
        dcm = getattr(orig_state, func)()
        assert dcm.get_state_dcm().shape == (6, 6)
        assert dcm.rot_mat.shape == (3, 3)
        assert dcm.rot_mat_dt.shape == (3, 3)
        print(f"== {func} ==\n{dcm}")
        # Test rebuilding the DCM from its parts
        dcm_rebuilt = DCM(dcm.rot_mat, dcm.from_id, dcm.to_id, dcm.rot_mat_dt)
        assert dcm_rebuilt == dcm

    # Check that we can convert a DCM to a quaternion and back
    q = dcm.to_quaternion()
    q_rebuilt = Quaternion(q.w, q.x, q.y, q.z, q.from_id, q.to_id)
    # Test quaternion multiplication
    conj_q = q_rebuilt.conjugate()
    assert (conj_q * q).is_zero()

    conj_dcm = conj_q.to_dcm()
    assert (conj_dcm * dcm).is_identity()

    uvec, angle = q.uvec_angle_rad()
    prv = q.prv()
    err = angle * uvec - prv
    assert sum([e**2 for e in err]) < 2e-16
    dcm_from_q = q.to_dcm()
    assert q.b_matrix().shape == (4, 3)

    orig_uvec, orig_angle_rad = q.uvec_angle_rad()
    rtn_uvec, rtn_angle_rad = dcm_from_q.to_quaternion().uvec_angle_rad()

    assert all(rtn_uvec == orig_uvec)
    assert rtn_angle_rad == orig_angle_rad

    topo_dcm = orig_state.dcm_from_topocentric_to_body_fixed()
    assert topo_dcm.get_state_dcm().shape == (6, 6)
    assert topo_dcm.rot_mat.shape == (3, 3)
    assert (
        topo_dcm.rot_mat_dt is not None and topo_dcm.rot_mat_dt.shape == (3, 3)
    ) or topo_dcm.rot_mat_dt is None

    # In Python, we can set the aberration to None
    aberration = None

    state_itrf93 = almanac.transform_to(orig_state, Frames.EARTH_ITRF93, aberration)

    print(orig_state)
    print(state_itrf93)

    assert abs(state_itrf93.latitude_deg() - 10.549246868302738) < 1e-10
    assert abs(state_itrf93.longitude_deg() - 133.76889100913047) < 1e-10
    assert abs(state_itrf93.height_km() - 1814.503598063825) < 1e-10

    # Convert back
    from_state_itrf93_to_eme2k = almanac.transform_to(state_itrf93, Frames.EARTH_J2000)

    print(from_state_itrf93_to_eme2k)

    assert orig_state == from_state_itrf93_to_eme2k

    # Demo creation of a ground station
    mean_earth_angular_velocity_deg_s = 0.004178079012116429
    # Grab the loaded frame info
    itrf93 = almanac.frame_info(Frames.EARTH_ITRF93)
    paris = Orbit.from_latlongalt(
        48.8566,
        2.3522,
        0.4,
        epoch,
        itrf93,
    )

    assert abs(paris.latitude_deg() - 48.8566) < 1e-3
    assert abs(paris.longitude_deg() - 2.3522) < 1e-3
    assert abs(paris.height_km() - 0.4) < 1e-3

    # Lat/long/alt high fidelity
    omega_itrf93 = almanac.angular_velocity_wrt_j2000_rad_s(Frames.EARTH_ITRF93, epoch)
    paris_prec = Orbit.from_latlongalt_omega(
        48.8566,
        2.3522,
        0.4,
        omega_itrf93,
        epoch,
        itrf93,
    )

    assert abs(paris_prec.latitude_deg() - 48.8566) < 1e-3
    assert abs(paris_prec.longitude_deg() - 2.3522) < 1e-3
    assert abs(paris_prec.height_km() - 0.4) < 1e-3
    # Test that the velocity for the high precision is correct.
    # Assume that the ITRF Z is the main axis of rotation, so there should not be much velocity there
    assert paris_prec.vz_km_s < 1e-3
    # Compute the perpendicular distance of the object
    r_perp_km = (paris_prec.x_km**2 + paris_prec.y_km**2) ** 0.5
    # Speed of that point based on its distance from the center of Earth
    speed_km_s = r_perp_km * radians(mean_earth_angular_velocity_deg_s)

    # Check the approximation of the mean angular velocity is correct.
    assert abs(speed_km_s - paris_prec.vmag_km_s()) < 1e-3

    # Pickling test
    pickle.loads(pickle.dumps(eme2k)) == eme2k
    pickle.loads(pickle.dumps(eme2k.shape)) == eme2k.shape
    # Cannot pickle across module boundaries =(
    # pickle.loads(pickle.dumps(paris)) == paris

    # Test that we can get the SPK data type
    assert int(almanac.spk_summaries(301)[0].datatype()) == 2, "not shown as type 2"

    # Function export test
    for fname in [
        "transform",
        "transform_to",
        "translate",
        "translate_to",
        "translate_geometric",
        "spk_ezr",
        "state_of",
        "solar_eclipsing",
        "occultation",
        "line_of_sight_obstructed",
        "azimuth_elevation_range_sez",
        "spk_domains",
        "spk_summaries",
    ]:
        assert hasattr(almanac, fname)

    # Test the parallel function calls
    start = Epoch("2021-10-29 12:34:56 TDB")
    stop = Epoch("2022-10-29 12:34:56 TDB")
    time_series = TimeSeries(
        start,
        stop,
        Duration("1 min"),
        False,
    )

    tick = Epoch.system_now()

    states = almanac.transform_many(
        Frames.EARTH_J2000,
        Frames.SUN_J2000,
        time_series,
        None,
    )

    clock_time = Epoch.system_now().timedelta(tick)
    print(f"Queried {len(states)} states in {clock_time}")
    assert len(states) == int(stop.timedelta(start).to_unit(Unit.Minute))


def test_numpy_constructor():
    if environ.get("CI", False):
        # Load from meta kernel to not use Git LFS quota
        data_path = Path(__file__).parent.joinpath(
            "..", "..", "data", "ci_config.dhall"
        )
        meta = MetaAlmanac(str(data_path))
        almanac = meta.process()
    else:
        data_path = Path(__file__).parent.joinpath("..", "..", "data")
        almanac = Almanac(str(data_path.joinpath("pck08.pca")))

    epoch = Epoch("2021-10-29 12:34:56 TDB")
    eme2k = almanac.frame_info(Frames.EME2000)

    state_vector = np.array([8_191.93, 0.0, 0.0, 0.0, 7.6, 0.0])

    orbit_from_npy = Orbit(state_vector, epoch, eme2k)

    orbit_from_floats = Orbit(8_191.93, 0.0, 0.0, 0.0, 7.6, 0.0, epoch, eme2k)

    assert orbit_from_npy == orbit_from_floats

    # Test that it fails with a wrong-sized array
    with np.testing.assert_raises(ValueError):
        Orbit(np.array([1.0, 2.0, 3.0]), epoch, eme2k)

    assert np.all(orbit_from_npy.radius_km() == np.array([8_191.93, 0.0, 0.0]))
    assert np.all(orbit_from_npy.velocity_km_s() == np.array([0.0, 7.6, 0.0]))


def test_convert_tpc():
    """Attempt to reproduce GH issue #339"""
    try:
        os.remove("test_constants.tpc")
    except FileNotFoundError:
        pass

    # First call to convert_tpc works
    convert_tpc("data/pck00011.tpc", "data/gm_de440.tpc", "test_constants.tpc")

    # Second call, with overwrite enabled, also works
    convert_tpc(
        "data/pck00011.tpc", "data/gm_de440.tpc", "test_constants.tpc", overwrite=True
    )

    # Try to load the constants file
    constants_file = MetaFile("test_constants.tpc")
    new_meta = MetaAlmanac()
    new_meta.files = [
        constants_file,
    ]
    almanac = new_meta.process()

    earth_j2k = almanac.frame_info(Frames.EARTH_J2000)
    assert earth_j2k.mu_km3_s2 is not None
    almanac.describe()


def test_meta_load():
    data_path = Path(__file__).parent.joinpath("..", "..", "data", "local.dhall")
    meta = MetaAlmanac(str(data_path))
    print(meta)
    try:
        # Process the files to be loaded
        almanac = meta.process()
    except Exception as e:
        print("Not sure where the files are on Github CI")
        print(e)
    else:
        # And check that everything is loaded
        eme2k = almanac.frame_info(Frames.EME2000)
        assert eme2k.mu_km3_s2() == 398600.435436096
        assert eme2k.shape.polar_radius_km == 6356.75
        assert abs(eme2k.shape.flattening() - 0.0033536422844278) < 2e-16


def test_exports():
    for cls in [Frame, Ellipsoid, Orbit]:
        print(f"{cls} OK")


def test_frame_defs():
    print(f"{Frames.SSB_J2000}")
    print(dir(Frames))
    assert Frames.EME2000 == Frames.EME2000
    assert Frames.EME2000 == Frames.EARTH_J2000
    assert Frames.EME2000 != Frames.SSB_J2000


def test_location():
    mask = [TerrainMask(0.0, 5.0), TerrainMask(35.0, 10.0), TerrainMask(270.0, 3.0)]
    dss65 = Location(
        40.427_222,
        4.250_556,
        0.834_939,
        FrameUid(399, 399),
        mask,
        terrain_mask_ignored=False,
    )

    as_dhall = dss65.to_dhall()

    from_dhall = Location.from_dhall(as_dhall)

    print(from_dhall)

    # To build a location data set kernel, we must first build a location dhall set entry
    entry = LocationDhallSetEntry(dss65, id=1, alias="My Alias")
    # Then we append it to a LocationDhallSet
    dhallset = LocationDhallSet([entry])
    assert "data" in dir(dhallset), "missing getting on dhall set"
    # Now, we can build the kernel
    dataset = dhallset.to_dataset()
    # Save it as a Location Kernel ANISE (LKA) file, overwritting it if it exists
    dataset.save_as("pytest_loc_kernel.lka", True)
    # Reload it
    reloaded = LocationDataSet.load("pytest_loc_kernel.lka")
    # We can also convert it its Dhall representation
    dhallset = reloaded.to_dhallset()
    print(dhallset.to_dhall())
    # Confirm that we can load it in the almanac.
    almanac = Almanac("pytest_loc_kernel.lka")
    # Recall: the describe has its own print!
    almanac.describe()
    # And we can grab the location data itself
    my_loc = dhallset.data[0]
    print(my_loc.alias)
    print(my_loc.value)  # This is the location info
    terrain_mask = my_loc.value.terrain_mask
    print(terrain_mask)


def test_version():
    from anise import __version__

    assert __version__ is not None


def test_oem():
    # Load an Almanac for the various frames used here.
    almanac = Almanac("data/pck11.pca")
    # 1. Load an OEM to Ephem
    ephem = Ephemeris.from_ccsds_oem_file("data/tests/ccsds/oem/LRO_Nyx.oem")
    print(ephem)
    assert ephem.degree == 7, f"Expected degree 7, got {ephem.degree}"
    (start, end) = ephem.domain()
    # Query the covariance using the Log Euclidean method
    covar = ephem.covar_at(start + (end - start) * 0.5, LocalFrame.RIC, almanac)
    print(covar)
    covar = ephem.covar_at(start + (end - start) * 0.5, LocalFrame.Inertial, almanac)
    print(covar)
    covar = ephem.covar_at(start + (end - start) * 0.5, LocalFrame.RCN, almanac)
    print(covar)
    covar = ephem.covar_at(start + (end - start) * 0.5, LocalFrame.VNC, almanac)
    print(covar)
    # Export to SPICE BSP
    ephem.write_spice_bsp(
        -159,
        "data/tests/naif/spk/ephem_from_python.bsp",
        DataType.Type13HermiteUnequalStep,
    )
    # Export to CCSDS OEM
    ephem.write_ccsds_oem(
        "data/tests/naif/spk/ephem_from_python.oem", "My Originator", "OBJECT_NAME"
    )
    # Ensure we can read what we wrote
    ephem_reread = Ephemeris.from_ccsds_oem_file(
        "data/tests/naif/spk/ephem_from_python.oem"
    )
    assert ephem_reread.start_epoch() == ephem.start_epoch()
    assert ephem_reread.end_epoch() == ephem.end_epoch()

    # 2. Load OEM to new Almanac, providing an ID to convert this OEM to SPK
    almanac2 = Almanac.from_ccsds_oem_file("data/tests/ccsds/oem/LRO_Nyx.oem", -159)
    start2, end2 = almanac2.spk_domain(-159)
    # Small difference due to Ephemeris Time conversion, cf. hifitime docs
    assert (start2 - start).abs().to_seconds() < 1e-7
    assert (end2 - end).abs().to_seconds() < 1e-7
    # 3. Load OEM to existing almanac
    almanac = almanac.load_ccsds_oem_file("data/tests/ccsds/oem/LRO_Nyx.oem", -160)
    start3, end3 = almanac.spk_domain(-160)
    assert start2 == start3
    assert end2 == end3
    print(end3.to_et_seconds())
    # Compute the standard deviation for the covariance
    sigma_sma_km = ephem.at(end, almanac).sigma_for(OrbitalElement.SemiMajorAxis)
    print(f"SMA 1-sigma = {sigma_sma_km:.3} km")

    # Build an Ephemeris
    eme2k = almanac.frame_info(Frames.EARTH_J2000)
    epoch = Epoch("2024-02-29T12:34:56")

    orbit = Orbit.from_keplerian_altitude(
        567.8, 1e-4, 28.5, 75.0, 115.0, 0.0, epoch, eme2k
    )

    ts = TimeSeries(epoch, epoch + Unit.Day * 1, Unit.Minute * 1, True)

    records = []
    ephem_from_empty = Ephemeris([], "My Spacecraft")

    for tickytack in ts:
        orbit_t = orbit.at_epoch(tickytack)
        cov = Covariance(np.diag([1e3] * 6), LocalFrame.RIC)
        record = EphemerisRecord(orbit_t, cov)
        records += [record]
        ephem_from_empty.insert_orbit(orbit_t)
    # Add an entry one tiny step later
    final_orbit = orbit.at_epoch(records[-1].orbit.epoch + Unit.Nanosecond * 4)
    ephem_from_empty.insert_orbit(final_orbit)

    ephem_from_list = Ephemeris([rcrd.orbit for rcrd in records], "My Spacecraft Too!")
    ephem_from_list.insert_orbit(final_orbit)

    assert ephem_from_list.start_epoch() == ephem_from_empty.start_epoch()
    assert ephem_from_list.end_epoch() == ephem_from_empty.end_epoch()
    assert ephem_from_list.len() == ephem_from_empty.len()

    # Check that we can iterate over an ephem entry by entry
    entry_count = 0
    for entry in ephem:
        entry_count += 1
        assert isinstance(entry, EphemerisRecord)
    assert entry_count == ephem.len()


if __name__ == "__main__":
    # test_meta_load()
    # test_exports()
    # test_frame_defs()
    # test_convert_tpc()
    # test_state_transformation()
    # test_location()
    test_oem()



import argparse
import ast
import importlib
import inspect
import logging
import re
import subprocess
from functools import reduce
from typing import Any, Dict, List, Mapping, Optional, Set, Tuple, Union


def path_to_type(*elements: str) -> ast.AST:
    base: ast.AST = ast.Name(id=elements[0], ctx=ast.Load())
    for e in elements[1:]:
        base = ast.Attribute(value=base, attr=e, ctx=ast.Load())
    return base


OBJECT_MEMBERS = dict(inspect.getmembers(object))
BUILTINS: Dict[str, Union[None, Tuple[List[ast.AST], ast.AST]]] = {
    "__annotations__": None,
    "__bool__": ([], path_to_type("bool")),
    "__bytes__": ([], path_to_type("bytes")),
    "__class__": None,
    "__contains__": ([path_to_type("typing", "Any")], path_to_type("bool")),
    "__del__": None,
    "__delattr__": ([path_to_type("str")], path_to_type("None")),
    "__delitem__": ([path_to_type("typing", "Any")], path_to_type("None")),
    "__dict__": None,
    "__dir__": None,
    "__doc__": None,
    "__eq__": ([path_to_type("typing", "Any")], path_to_type("bool")),
    "__format__": ([path_to_type("str")], path_to_type("str")),
    "__ge__": ([path_to_type("typing", "Any")], path_to_type("bool")),
    "__getattribute__": ([path_to_type("str")], path_to_type("typing", "Any")),
    "__getitem__": ([path_to_type("typing", "Any")], path_to_type("typing", "Any")),
    "__gt__": ([path_to_type("typing", "Any")], path_to_type("bool")),
    "__hash__": ([], path_to_type("int")),
    "__init__": ([], path_to_type("None")),
    "__init_subclass__": None,
    "__iter__": ([], path_to_type("typing", "Any")),
    "__le__": ([path_to_type("typing", "Any")], path_to_type("bool")),
    "__len__": ([], path_to_type("int")),
    "__lt__": ([path_to_type("typing", "Any")], path_to_type("bool")),
    "__module__": None,
    "__ne__": ([path_to_type("typing", "Any")], path_to_type("bool")),
    "__new__": None,
    "__next__": ([], path_to_type("typing", "Any")),
    "__int__": ([], path_to_type("None")),
    "__reduce__": None,
    "__reduce_ex__": None,
    "__repr__": ([], path_to_type("str")),
    "__setattr__": (
        [path_to_type("str"), path_to_type("typing", "Any")],
        path_to_type("None"),
    ),
    "__setitem__": (
        [path_to_type("typing", "Any"), path_to_type("typing", "Any")],
        path_to_type("None"),
    ),
    "__sizeof__": None,
    "__str__": ([], path_to_type("str")),
    "__subclasshook__": None,
    "__add__": ([path_to_type("typing", "Any")], path_to_type("typing", "Any")),
    "__sub__": ([path_to_type("typing", "Any")], path_to_type("typing", "Any")),
    "__mul__": ([path_to_type("typing", "Any")], path_to_type("typing", "Any")),
    "__truediv__": ([path_to_type("typing", "Any")], path_to_type("typing", "Any")),
    "__div__": ([path_to_type("typing", "Any")], path_to_type("typing", "Any")),
    "__radd__": ([path_to_type("typing", "Any")], path_to_type("typing", "Any")),
    "__rsub__": ([path_to_type("typing", "Any")], path_to_type("typing", "Any")),
    "__rmul__": ([path_to_type("typing", "Any")], path_to_type("typing", "Any")),
    "__rtruediv__": ([path_to_type("typing", "Any")], path_to_type("typing", "Any")),
    "__setstate__": ([path_to_type("typing", "Any")], path_to_type("None")),
    "__getnewargs__": ([], path_to_type("typing", "Tuple")),
}

TYPE_MAPPING = {
    # anise.time
    "Epoch": "anise.time.Epoch",
    "Duration": "anise.time.Duration",
    "TimeSeries": "anise.time.TimeSeries",
    "TimeScale": "anise.time.TimeScale",
    "Unit": "anise.time.Unit",
    "MonthName": "anise.time.MonthName",
    "Weekday": "anise.time.Weekday",
    # anise.astro
    "Frame": "anise.astro.Frame",
    "Orbit": "anise.astro.Orbit",
    "Location": "anise.astro.Location",
    "AzElRange": "anise.astro.AzElRange",
    "Occultation": "anise.astro.Occultation",
    "Ellipsoid": "anise.astro.Ellipsoid",
    "Ephemeris": "anise.astro.Ephemeris",
    "EphemerisRecord": "anise.astro.EphemerisRecord",
    "Mass": "anise.astro.Mass",
    "DragData": "anise.astro.DragData",
    "SRPData": "anise.astro.SRPData",
    "DafDataType": "anise.astro.DafDataType",
    "LocalFrame": "anise.astro.LocalFrame",
    "Covariance": "anise.astro.Covariance",
    "TerrainMask": "anise.astro.TerrainMask",
    "FrameUid": "anise.astro.FrameUid",
    # anise (root)
    "Aberration": "anise.Aberration",
    "Almanac": "anise.Almanac",
    "MetaAlmanac": "anise.MetaAlmanac",
    "MetaFile": "anise.MetaFile",
    "LocationDataSet": "anise.LocationDataSet",
    "LocationDhallSet": "anise.LocationDhallSet",
    "LocationDhallSetEntry": "anise.LocationDhallSetEntry",
    "PyLocationDataSet": "anise.LocationDataSet",
    "PyReportScalars": "anise.analysis.ReportScalars",
    "ReportScalars": "anise.analysis.ReportScalars",
    # anise.rotation
    "Quaternion": "anise.rotation.Quaternion",
    "DCM": "anise.rotation.DCM",
    # anise.analysis
    "Condition": "anise.analysis.Condition",
    "Event": "anise.analysis.Event",
    "EventDetails": "anise.analysis.EventDetails",
    "EventEdge": "anise.analysis.EventEdge",
    "OrbitalElement": "anise.analysis.OrbitalElement",
    "Plane": "anise.analysis.Plane",
    "VisibilityArc": "anise.analysis.VisibilityArc",
    "StateSpec": "anise.analysis.StateSpec",
    "PyStateSpec": "anise.analysis.StateSpec",
    "EventArc": "anise.analysis.EventArc",
    "PyEventArc": "anise.analysis.EventArc",
    "FrameSpec": "anise.analysis.FrameSpec",
    "PyFrameSpec": "anise.analysis.FrameSpec",
    "OrthogonalFrame": "anise.analysis.OrthogonalFrame",
    "PyOrthogonalFrame": "anise.analysis.OrthogonalFrame",
    "ScalarExpr": "anise.analysis.ScalarExpr",
    "PyScalarExpr": "anise.analysis.ScalarExpr",
    "VectorExpr": "anise.analysis.VectorExpr",
    "PyVectorExpr": "anise.analysis.VectorExpr",
    # anise.instrument
    "Instrument": "anise.instrument.Instrument",
    "FovShape": "anise.instrument.FovShape",
    # third-party
    "numpy.array": "numpy.ndarray",
    "numpy.ndarray": "numpy.ndarray",
    "np.ndarray": "numpy.ndarray",
    "np.array": "numpy.ndarray",
    "tuples": "typing.Tuple",
}

def module_stubs(module: Any) -> ast.Module:
    types_to_import = {"typing"}
    classes = []
    functions = []
    for member_name, member_value in inspect.getmembers(module):
        element_path = [module.__name__, member_name]
        if member_name.startswith("__"):
            pass
        elif inspect.isclass(member_value):
            classes.append(
                class_stubs(member_name, member_value, element_path, types_to_import)
            )
        elif inspect.isbuiltin(member_value) or inspect.isroutine(member_value):
            functions.append(
                function_stub(
                    member_name,
                    member_value,
                    element_path,
                    types_to_import,
                    in_class=False,
                )
            )
        else:
            logging.warning(f"Unsupported root construction {member_name}")
    # Add metadata constants if they exist in the module
    for meta in ["__author__", "__version__"]:
        if hasattr(module, meta):
            functions.append(
                ast.AnnAssign(
                    target=ast.Name(id=meta, ctx=ast.Store()),
                    annotation=ast.Name(id="str", ctx=ast.Load()),
                    value=ast.Constant(getattr(module, meta)),
                    simple=1,
                )
            )

    # Resolve all types to import
    imports = [
        ast.ImportFrom(module="__future__", names=[ast.alias(name="annotations")], level=0)
    ]
    for t in sorted(types_to_import):
        if t == "typing":
            imports.append(ast.Import(names=[ast.alias(name=t)]))
            continue

        if t == "numpy":
             imports.append(ast.Import(names=[ast.alias(name="numpy")]))
             continue

        if t.startswith("anise."):
            current_module = module.__name__
            if current_module == t:
                 continue

            parts = t.split(".")
            # e.g. from anise import time
            imports.append(ast.ImportFrom(module=parts[0], names=[ast.alias(name=parts[1])], level=0))
        elif t == "anise":
             # Avoid self-import
             if module.__name__ != "anise":
                  imports.append(ast.Import(names=[ast.alias(name="anise")]))
        else:
            imports.append(ast.Import(names=[ast.alias(name=t)]))

    return ast.Module(
        body=imports + classes + functions,
        type_ignores=[],
    )


def class_stubs(
    cls_name: str, cls_def: Any, element_path: List[str], types_to_import: Set[str]
) -> ast.ClassDef:
    attributes: List[ast.AST] = []
    methods: List[ast.AST] = []
    magic_methods: List[ast.AST] = []
    constants: List[ast.AST] = []

    for member_name, member_value in inspect.getmembers(cls_def):
        current_element_path = [*element_path, member_name]
        # Inside the loop in class_stubs
        if member_name in ("__init__", "__new__") and "Error" not in cls_name:
            try:
                actual_member = cls_def if member_name == "__new__" else member_value
                methods.append(
                    function_stub(
                        member_name,
                        actual_member,
                        current_element_path,
                        types_to_import,
                        in_class=True,
                        cls_def=cls_def
                    )
                )
            except Exception:
                continue
        elif (
            member_value == OBJECT_MEMBERS.get(member_name)
            or BUILTINS.get(member_name, ()) is None
        ):
            pass
        elif inspect.isdatadescriptor(member_value):
            attributes.extend(
                data_descriptor_stub(
                    member_name, member_value, current_element_path, types_to_import
                )
            )
        elif inspect.isroutine(member_value) or inspect.isbuiltin(member_value):
            (magic_methods if member_name.startswith("__") else methods).append(
                function_stub(
                    member_name,
                    member_value,
                    current_element_path,
                    types_to_import,
                    in_class=True,
                    cls_def=cls_def
                )
            )
        elif member_name == "__match_args__":
            constants.append(
                ast.AnnAssign(
                    target=ast.Name(id=member_name, ctx=ast.Store()),
                    annotation=ast.Subscript(
                        value=path_to_type("tuple"),
                        slice=ast.Tuple(
                            elts=[path_to_type("str"), ast.Ellipsis()], ctx=ast.Load()
                        ),
                        ctx=ast.Load(),
                    ),
                    value=ast.Constant(member_value),
                    simple=1,
                )
            )
        elif member_value is not None and not member_name.startswith("_"):
            try:
                annotation = concatenated_path_to_type(
                        member_value.__class__.__name__, element_path, types_to_import
                    )
                constants.append(
                    ast.AnnAssign(
                        target=ast.Name(id=member_name, ctx=ast.Store()),
                        annotation=annotation,
                        value=ast.Ellipsis(),
                        simple=1,
                    )
                )
            except Exception:
                pass
        else:
            pass

    doc = inspect.getdoc(cls_def)
    doc_comment = build_doc_comment(doc) if doc else None
    return ast.ClassDef(
        cls_name,
        bases=[],
        keywords=[],
        body=(
            ([doc_comment] if doc_comment else [])
            + attributes
            + methods
            + magic_methods
            + constants
        )
        or [ast.Ellipsis()],
        decorator_list=[path_to_type("typing", "final")],
    )


def data_descriptor_stub(
    data_desc_name: str,
    data_desc_def: Any,
    element_path: List[str],
    types_to_import: Set[str],
) -> Union[Tuple[ast.AnnAssign, ast.Expr], Tuple[ast.AnnAssign]]:
    annotation = None
    doc_comment = None

    doc = inspect.getdoc(data_desc_def)
    if doc is not None:
        try:
            annotation = returns_stub(data_desc_name, doc, element_path, types_to_import)
        except Exception:
            pass
        m = re.findall(r"^ *:return: *(.*) *$", doc, re.MULTILINE)
        if len(m) == 1:
            doc_comment = m[0]
        elif len(m) > 1:
            raise ValueError(
                f"Multiple return annotations found with :return: in {'.'.join(element_path)} documentation"
            )

    assign = ast.AnnAssign(
        target=ast.Name(id=data_desc_name, ctx=ast.Store()),
        annotation=annotation or path_to_type("typing", "Any"),
        simple=1,
    )
    doc_comment = build_doc_comment(doc_comment) if doc_comment else None
    return (assign, doc_comment) if doc_comment else (assign,)


def function_stub(
    fn_name: str,
    fn_def: Any,
    element_path: List[str],
    types_to_import: Set[str],
    *,
    in_class: bool,
    cls_def: Any = None
) -> ast.FunctionDef:
    body: List[ast.AST] = []
    doc = inspect.getdoc(fn_def)
    if doc is not None:
        doc_comment = build_doc_comment(doc)
        if doc_comment is not None:
            body.append(doc_comment)

    decorator_list = []
    is_static = False
    is_class = False

    if in_class:
        try:
            # Check if it's explicitly classmethod or staticmethod
            actual_fn = getattr(cls_def, fn_name, None)
            if isinstance(actual_fn, classmethod):
                 decorator_list.append(ast.Name(id="classmethod", ctx=ast.Load()))
                 is_class = True
            elif isinstance(actual_fn, staticmethod):
                 decorator_list.append(ast.Name(id="staticmethod", ctx=ast.Load()))
                 is_static = True
            else:
                try:
                    sig = inspect.signature(fn_def)
                    params = list(sig.parameters.values())
                    if params and params[0].name == "cls":
                        decorator_list.append(ast.Name(id="classmethod", ctx=ast.Load()))
                        is_class = True
                    elif (
                        in_class
                        and hasattr(fn_def, "__self__")
                        and not isinstance(fn_def.__self__, type)
                    ):
                        # Standard instance method bound
                        pass
                    elif in_class and hasattr(fn_def, "__self__"):
                        decorator_list.append(ast.Name(id="staticmethod", ctx=ast.Load()))
                        is_static = True
                except ValueError:
                    # Fallback for builtins in exception classes
                    if "Error" in element_path[-2]:
                         pass # default to instance method
        except Exception:
            pass

    print(f"Documenting {fn_name}")

    return ast.FunctionDef(
        fn_name,
        arguments_stub(fn_name, fn_def, doc or "", element_path, types_to_import, in_class, is_static, is_class),
        body or [ast.Ellipsis()],
        decorator_list=decorator_list,
        returns=(
            returns_stub(fn_name, doc or "", element_path, types_to_import, in_class)
        ),
        lineno=0,
    )


def arguments_stub(
    callable_name: str,
    callable_def: Any,
    doc: str,
    element_path: List[str],
    types_to_import: Set[str],
    in_class: bool,
    is_static: bool,
    is_class: bool,
) -> ast.arguments:
    if "Error" in element_path[1]:
        # Don't document errors
        return ast.arguments(posonlyargs=[], args=[], defaults=[], kwonlyargs=[])

    builtin = BUILTINS.get(callable_name)
    try:
        sig = inspect.signature(callable_def)
        real_parameters: Mapping[str, inspect.Parameter] = sig.parameters
    except Exception:
        # Fallback for builtins without signatures
        args = []
        if in_class:
            if is_static:
                 pass
            elif callable_name == "__new__" or is_class:
                args.append(ast.arg(arg="cls", annotation=None))
            else:
                args.append(ast.arg(arg="self", annotation=None))

        # If we have a BUILTIN entry, use it for arguments too
        if isinstance(builtin, tuple):
             for i, t in enumerate(builtin[0]):
                  args.append(ast.arg(arg=f"arg{i}", annotation=t))
             return ast.arguments(posonlyargs=[], args=args, defaults=[], kwonlyargs=[])

        return ast.arguments(
            posonlyargs=[],
            args=args,
            vararg=ast.arg(arg="args", annotation=path_to_type("typing", "Any")),
            kwonlyargs=[],
            kw_defaults=[],
            defaults=[],
            kwarg=ast.arg(arg="kwargs", annotation=path_to_type("typing", "Any")),
        )

    modified_parameters = dict(real_parameters)
    if in_class:
        if callable_name == "__init__":
            if "self" not in modified_parameters:
                modified_parameters = {
                    "self": inspect.Parameter("self", inspect.Parameter.POSITIONAL_ONLY),
                    **modified_parameters,
                }
        elif callable_name == "__new__":
            if "cls" not in modified_parameters:
                modified_parameters = {
                    "cls": inspect.Parameter("cls", inspect.Parameter.POSITIONAL_ONLY),
                    **modified_parameters,
                }
        elif not is_static and not any(p.name in ("self", "cls") for p in real_parameters.values()):
             if is_class:
                  modified_parameters = {
                    "cls": inspect.Parameter("cls", inspect.Parameter.POSITIONAL_ONLY),
                    **modified_parameters,
                }
             else:
                  modified_parameters = {
                    "self": inspect.Parameter("self", inspect.Parameter.POSITIONAL_ONLY),
                    **modified_parameters,
                }

    parsed_param_types = {}
    optional_params = set()

    # Types for magic functions types
    if isinstance(builtin, tuple):
        param_names = list(modified_parameters.keys())
        if param_names and param_names[0] in ("self", "cls"):
            del param_names[0]
        for name, t in zip(param_names, builtin[0]):
            parsed_param_types[name] = t

    elif callable_name in [
        "__add__",
        "__sub__",
        "__div__",
        "__mul__",
        "__radd__",
        "__rsub__",
        "__rdiv__",
        "__rmul__",
    ]:
        return ast.arguments(posonlyargs=[], args=[], defaults=[], kwonlyargs=[])

    # Types from comment
    for match in re.findall(
        r"^ *:type *([a-zA-Z0-9_]+): ([^\n]*) *$", doc, re.MULTILINE
    ):
        if match[0] not in modified_parameters:
            logging.warning(
                f"The parameter {match[0]} of {'.'.join(element_path)} "
                "is defined in the documentation but not in the function signature"
            )
            continue
        type_str = match[1]
        if type_str.endswith(", optional"):
            optional_params.add(match[0])
            type_str = type_str[:-10]
        parsed_param_types[match[0]] = convert_type_from_doc(
            type_str, element_path, types_to_import
        )

    # we parse the parameters
    posonlyargs = []
    args = []
    vararg = None
    kwonlyargs = []
    kw_defaults = []
    kwarg = None
    defaults = []
    for param in modified_parameters.values():
        if param.name in ("self", "cls"):
            param_ast = ast.arg(arg=param.name, annotation=None)
        elif param.name not in parsed_param_types:
            logging.warning(
                f"The parameter {param.name} of {'.'.join(element_path)} "
                "has no type definition in the function documentation"
            )
            param_ast = ast.arg(arg=param.name, annotation=path_to_type("typing", "Any"))
        else:
            annotation = parsed_param_types.get(param.name)
            if param.name in optional_params or param.default != param.empty:
                annotation = ast.Subscript(
                    value=path_to_type("typing", "Optional"),
                    slice=annotation,
                    ctx=ast.Load()
                )
                types_to_import.add("typing")
            param_ast = ast.arg(arg=param.name, annotation=annotation)

        default_ast = None
        if param.default != param.empty:
            default_ast = ast.Constant(param.default)

        if param.kind == param.POSITIONAL_ONLY:
            args.append(param_ast)
        elif param.kind == param.POSITIONAL_OR_KEYWORD:
            args.append(param_ast)
            defaults.append(default_ast)
        elif param.kind == param.VAR_POSITIONAL:
            vararg = param_ast
        elif param.kind == param.KEYWORD_ONLY:
            kwonlyargs.append(param_ast)
            kw_defaults.append(default_ast)
        elif param.kind == param.VAR_KEYWORD:
            kwarg = param_ast

    return ast.arguments(
        posonlyargs=posonlyargs,
        args=args,
        vararg=vararg,
        kwonlyargs=kwonlyargs,
        kw_defaults=kw_defaults,
        defaults=defaults,
        kwarg=kwarg,
    )


def returns_stub(
    callable_name: str, doc: str, element_path: List[str], types_to_import: Set[str], in_class: bool = False
) -> Optional[ast.AST]:
    if "Error" in element_path[1]:
        # Don't document errors
        return

    if callable_name in [
        "__add__",
        "__sub__",
        "__div__",
        "__mul__",
        "__radd__",
        "__rsub__",
        "__rdiv__",
        "__rmul__",
    ]:
        return

    if callable_name == "__new__":
         # Returns the class itself
         if len(element_path) >= 2:
              return path_to_type(element_path[-2])
         return path_to_type("typing", "Any")

    m = re.findall(r"^ *:rtype: *([^\n]*) *$", doc, re.MULTILINE)
    if len(m) == 0:
        builtin = BUILTINS.get(callable_name)
        if isinstance(builtin, tuple) and builtin[1] is not None:
            return builtin[1]
        return path_to_type("typing", "Any")
    if len(m) > 1:
        raise ValueError(
            f"Multiple return type annotations found with :rtype: for {'.'.join(element_path)}"
        )
    return convert_type_from_doc(m[0], element_path, types_to_import)


def convert_type_from_doc(
    type_str: str, element_path: List[str], types_to_import: Set[str]
) -> ast.AST:
    type_str = type_str.strip()
    return parse_type_to_ast(type_str, element_path, types_to_import)


def parse_type_to_ast(
    type_str: str, element_path: List[str], types_to_import: Set[str]
) -> ast.AST:
    # Resolve known types
    if type_str in TYPE_MAPPING:
        type_str = TYPE_MAPPING[type_str]

    # If it's a type in the current module, we can strip the module name
    current_module = element_path[0]
    if type_str.startswith(current_module + "."):
        remainder = type_str[len(current_module)+1:]
        if "." not in remainder:
             type_str = remainder

    # let's tokenize
    tokens = []
    current_token = ""
    for c in type_str:
        if (
            "a" <= c <= "z"
            or "A" <= c <= "Z"
            or "0" <= c <= "9"
            or c == "_"
            or c == "."
        ):
            current_token += c
        else:
            if current_token:
                tokens.append(current_token)
            current_token = ""
            if c != " ":
                tokens.append(c)
    if current_token:
        tokens.append(current_token)

    # let's first parse nested parenthesis
    stack: List[List[Any]] = [[]]
    for token in tokens:
        if token == "[":
            children: List[str] = []
            stack[-1].append(children)
            stack.append(children)
        elif token == "]":
            stack.pop()
        else:
            stack[-1].append(token)

    def parse_sequence(sequence: List[Any]) -> ast.AST:
        # 1. Handle commas first: split the sequence into distinct arguments
        args: List[List[Any]] = [[]]
        for e in sequence:
            if e == ",":
                args.append([])
            else:
                args[-1].append(e)

        # Filter out empty args (e.g. trailing commas)
        args = [a for a in args if a]

        # 2. If there are multiple arguments (separated by commas),
        # we return an ast.Tuple (this handles tuple[int, int, ...])
        if len(args) > 1:
            return ast.Tuple(elts=[parse_sequence(arg) for arg in args], ctx=ast.Load())

        # 3. Existing logic for "or" (Union types) within a single argument
        actual_sequence = args[0]
        or_groups: List[List[Any]] = [[]]
        for e in actual_sequence:
            if e == "or":
                or_groups.append([])
            else:
                or_groups[-1].append(e)

        new_elements: List[ast.AST] = []
        for group in or_groups:
            if len(group) == 1 and isinstance(group[0], str):
                # Standard type: int
                new_elements.append(
                    concatenated_path_to_type(group[0], element_path, types_to_import)
                )
            elif (
                len(group) == 2
                and isinstance(group[0], str)
                and isinstance(group[1], list)
            ):
                # Nested type: list[int] or tuple[...]
                new_elements.append(
                    ast.Subscript(
                        value=concatenated_path_to_type(
                            group[0], element_path, types_to_import
                        ),
                        slice=parse_sequence(group[1]),
                        ctx=ast.Load(),
                    )
                )
            else:
                raise ValueError(
                    f"Not able to parse type fragment '{group}' used by {'.'.join(element_path)}"
                )

        if len(new_elements) == 1:
            return new_elements[0]

        return reduce(
            lambda left, right: ast.BinOp(left=left, op=ast.BitOr(), right=right),
            new_elements,
        )

    return parse_sequence(stack[0])


def concatenated_path_to_type(
    path: str, element_path: List[str], types_to_import: Set[str]
) -> ast.AST:
    # Resolve known types
    if path in TYPE_MAPPING:
        path = TYPE_MAPPING[path]

    if path == "numpy.ndarray":
         types_to_import.add("numpy")
         return path_to_type("numpy", "ndarray")

    parts = path.split(".")

    current_module = element_path[0]

    if path.startswith("anise."):
         # Check if it's in our module
         if path.startswith(current_module + "."):
              remainder = path[len(current_module)+1:]
              if "." not in remainder:
                   # Local type
                   return path_to_type(remainder)

         subparts = path.split(".")
         if len(subparts) >= 2:
              # from anise import time -> time.Epoch
              types_to_import.add(".".join(subparts[:2]))
              return path_to_type(*subparts[1:])

    # If it's a type in the current module, we can strip the module name
    if len(parts) > 1:
        current_module_parts = current_module.split(".")
        if parts[:len(current_module_parts)] == current_module_parts:
            parts = parts[len(current_module_parts):]
            if len(parts) == 1:
                 return path_to_type(*parts)
        else:
            types_to_import.add(".".join(parts[:-1]))

    return path_to_type(*parts)


def build_doc_comment(doc: str) -> Optional[ast.Expr]:
    lines = [line.strip() for line in doc.split("\n")]
    clean_lines = []
    for line in lines:
        if line.startswith((":type", ":rtype")):
            continue
        clean_lines.append(line)
    text = "\n".join(clean_lines).strip()
    return ast.Expr(value=ast.Constant(text)) if text else None


def format_with_ruff(file: str) -> None:
    subprocess.check_call(["python", "-m", "ruff", "format", file])


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Extract Python type stub from a python module."
    )
    parser.add_argument(
        "module_name", help="Name of the Python module for which generate stubs"
    )
    parser.add_argument(
        "out",
        help="Name of the Python stub file to write to",
        type=argparse.FileType("wt"),
    )
    parser.add_argument(
        "--ruff", help="Formats the generated stubs using Ruff", action="store_true"
    )
    args = parser.parse_args()
    stub_content = ast.unparse(module_stubs(importlib.import_module(args.module_name)))
    args.out.write(stub_content)
    if args.ruff:
        format_with_ruff(args.out.name)



from __future__ import annotations
from anise import astro
import typing

@typing.final
class CelestialObjects:

    def __init__(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        """Initialize self.  See help(type(self)) for accurate signature."""

    def __new__(cls):...
    EARTH: int = ...
    EARTH_MOON_BARYCENTER: int = ...
    JUPITER: int = ...
    JUPITER_BARYCENTER: int = ...
    MARS: int = ...
    MARS_BARYCENTER: int = ...
    MERCURY: int = ...
    MOON: int = ...
    NEPTUNE: int = ...
    NEPTUNE_BARYCENTER: int = ...
    PLUTO_BARYCENTER: int = ...
    SATURN: int = ...
    SATURN_BARYCENTER: int = ...
    SOLAR_SYSTEM_BARYCENTER: int = ...
    SUN: int = ...
    URANUS: int = ...
    URANUS_BARYCENTER: int = ...
    VENUS: int = ...

@typing.final
class Frames:

    def __init__(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        """Initialize self.  See help(type(self)) for accurate signature."""

    def __new__(cls):...
    EARTH_ECLIPJ2000: astro.Frame = ...
    EARTH_ITRF93: astro.Frame = ...
    EARTH_J2000: astro.Frame = ...
    EARTH_MOON_BARYCENTER_J2000: astro.Frame = ...
    EME2000: astro.Frame = ...
    IAU_EARTH_FRAME: astro.Frame = ...
    IAU_JUPITER_FRAME: astro.Frame = ...
    IAU_MARS_FRAME: astro.Frame = ...
    IAU_MERCURY_FRAME: astro.Frame = ...
    IAU_MOON_FRAME: astro.Frame = ...
    IAU_NEPTUNE_FRAME: astro.Frame = ...
    IAU_SATURN_FRAME: astro.Frame = ...
    IAU_URANUS_FRAME: astro.Frame = ...
    IAU_VENUS_FRAME: astro.Frame = ...
    JUPITER_BARYCENTER_J2000: astro.Frame = ...
    MARS_BARYCENTER_J2000: astro.Frame = ...
    MERCURY_J2000: astro.Frame = ...
    MOON_J2000: astro.Frame = ...
    MOON_ME_DE421_FRAME: astro.Frame = ...
    MOON_ME_DE440_ME421_FRAME: astro.Frame = ...
    MOON_ME_FRAME: astro.Frame = ...
    MOON_PA_DE421_FRAME: astro.Frame = ...
    MOON_PA_DE440_FRAME: astro.Frame = ...
    MOON_PA_FRAME: astro.Frame = ...
    NEPTUNE_BARYCENTER_J2000: astro.Frame = ...
    PLUTO_BARYCENTER_J2000: astro.Frame = ...
    SATURN_BARYCENTER_J2000: astro.Frame = ...
    SSB_J2000: astro.Frame = ...
    SUN_J2000: astro.Frame = ...
    URANUS_BARYCENTER_J2000: astro.Frame = ...
    VENUS_J2000: astro.Frame = ...

@typing.final
class Orientations:

    def __init__(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        """Initialize self.  See help(type(self)) for accurate signature."""

    def __new__(cls):...
    ECLIPJ2000: int = ...
    IAU_EARTH: int = ...
    IAU_JUPITER: int = ...
    IAU_MARS: int = ...
    IAU_MERCURY: int = ...
    IAU_MOON: int = ...
    IAU_NEPTUNE: int = ...
    IAU_SATURN: int = ...
    IAU_URANUS: int = ...
    IAU_VENUS: int = ...
    ITRF93: int = ...
    J2000: int = ...
    MOON_ME: int = ...
    MOON_ME_DE421: int = ...
    MOON_ME_DE440_ME421: int = ...
    MOON_PA: int = ...
    MOON_PA_DE421: int = ...
    MOON_PA_DE440: int = ...

@typing.final
class UsualConstants:

    def __init__(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        """Initialize self.  See help(type(self)) for accurate signature."""

    def __new__(cls):...
    MEAN_EARTH_ANGULAR_VELOCITY_DEG_S: float = ...
    MEAN_MOON_ANGULAR_VELOCITY_DEG_S: float = ...
    SPEED_OF_LIGHT_KM_S: float = ...


from __future__ import annotations
from anise import analysis
from anise import astro
from anise import rotation
from anise import time
import numpy
import typing

@typing.final
class Aberration:
    """Represents the aberration correction options in ANISE.

    In space science and engineering, accurately pointing instruments (like optical cameras or radio antennas) at a target is crucial. This task is complicated by the finite speed of light, necessitating corrections for the apparent position of the target.

    This structure holds parameters for aberration corrections applied to a target's position or state vector. These corrections account for the difference between the target's geometric (true) position and its apparent position as observed.

    # Rule of tumb
    In most Earth orbits, one does _not_ need to provide any aberration corrections. Light time to the target is less than one second (the Moon is about one second away).
    In near Earth orbits, e.g. inner solar system, preliminary analysis can benefit from enabling unconverged light time correction. Stellar aberration is probably not required.
    For deep space missions, preliminary analysis would likely require both light time correction and stellar aberration. Mission planning and operations will definitely need converged light-time calculations.

    For more details, <https://naif.jpl.nasa.gov/pub/naif/toolkit_docs/C/req/abcorr.html>.

    # SPICE Validation

    The validation test `validate_jplde_de440s_aberration_lt` checks 101,000 pairs of ephemeris computations and shows that the unconverged Light Time computation matches the SPICE computations almost all the time.
    More specifically, the 99th percentile of error is less than 5 meters, the 75th percentile is less than one meter, and the median error is less than 2 millimeters."""

    converged: bool
    stellar: bool
    transmit_mode: bool

    def __init__(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        """Initialize self.  See help(type(self)) for accurate signature."""

    def __new__(cls, name: str) -> Aberration:
        """Represents the aberration correction options in ANISE.

        In space science and engineering, accurately pointing instruments (like optical cameras or radio antennas) at a target is crucial. This task is complicated by the finite speed of light, necessitating corrections for the apparent position of the target.

        This structure holds parameters for aberration corrections applied to a target's position or state vector. These corrections account for the difference between the target's geometric (true) position and its apparent position as observed.

        # Rule of tumb
        In most Earth orbits, one does _not_ need to provide any aberration corrections. Light time to the target is less than one second (the Moon is about one second away).
        In near Earth orbits, e.g. inner solar system, preliminary analysis can benefit from enabling unconverged light time correction. Stellar aberration is probably not required.
        For deep space missions, preliminary analysis would likely require both light time correction and stellar aberration. Mission planning and operations will definitely need converged light-time calculations.

        For more details, <https://naif.jpl.nasa.gov/pub/naif/toolkit_docs/C/req/abcorr.html>.

        # SPICE Validation

        The validation test `validate_jplde_de440s_aberration_lt` checks 101,000 pairs of ephemeris computations and shows that the unconverged Light Time computation matches the SPICE computations almost all the time.
        More specifically, the 99th percentile of error is less than 5 meters, the 75th percentile is less than one meter, and the median error is less than 2 millimeters."""

    def __eq__(self, value: typing.Any) -> bool:
        """Return self==value."""

    def __ge__(self, value: typing.Any) -> bool:
        """Return self>=value."""

    def __gt__(self, value: typing.Any) -> bool:
        """Return self>value."""

    def __le__(self, value: typing.Any) -> bool:
        """Return self<=value."""

    def __lt__(self, value: typing.Any) -> bool:
        """Return self<value."""

    def __ne__(self, value: typing.Any) -> bool:
        """Return self!=value."""

    def __repr__(self) -> str:
        """Return repr(self)."""

    def __str__(self) -> str:
        """Return str(self)."""

@typing.final
class Almanac:
    """An Almanac contains all of the loaded SPICE and ANISE data. It is the context for all computations."""

    def __init__(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        """Initialize self.  See help(type(self)) for accurate signature."""

    def __new__(cls, path: str) -> Almanac:
        """An Almanac contains all of the loaded SPICE and ANISE data. It is the context for all computations."""

    def angular_velocity_deg_s(
        self, from_frame: astro.Frame, to_frame: astro.Frame, epoch: time.Epoch
    ) -> numpy.ndarray:
        """Returns the angular velocity vector in deg/s of the from_frame wrt to the to_frame.

        This can be used to compute the angular velocity of the Earth ITRF93 frame with respect to the J2000 frame for example."""

    def angular_velocity_rad_s(
        self, from_frame: astro.Frame, to_frame: astro.Frame, epoch: time.Epoch
    ) -> numpy.ndarray:
        """Returns the angular velocity vector in rad/s of the from_frame wrt to the to_frame.

        This can be used to compute the angular velocity of the Earth ITRF93 frame with respect to the J2000 frame for example."""

    def angular_velocity_wrt_j2000_deg_s(
        self, from_frame: astro.Frame, epoch: time.Epoch
    ) -> numpy.ndarray:
        """Returns the angular velocity vector in deg/s of the from_frame wrt to the J2000 frame."""

    def angular_velocity_wrt_j2000_rad_s(
        self, from_frame: astro.Frame, epoch: time.Epoch
    ) -> numpy.ndarray:
        """Returns the angular velocity vector in rad/s of the from_frame wrt to the J2000 frame."""

    def azimuth_elevation_range_sez(
        self,
        rx: astro.Orbit,
        tx: astro.Orbit,
        obstructing_body: typing.Optional[astro.Frame] = None,
        ab_corr: typing.Optional[Aberration] = None,
    ) -> astro.AzElRange:
        """Computes the azimuth (in degrees), elevation (in degrees), and range (in kilometers) of the
        receiver state (`rx`) seen from the transmitter state (`tx`), once converted into the SEZ frame of the transmitter.

        # Warning
        The obstructing body _should_ be a tri-axial ellipsoid body, e.g. IAU_MOON_FRAME.

        # Algorithm
        1. If any obstructing_bodies are provided, ensure that none of these are obstructing the line of sight between the receiver and transmitter.
        2. Compute the SEZ (South East Zenith) frame of the transmitter.
        3. Rotate the receiver position vector into the transmitter SEZ frame.
        4. Rotate the transmitter position vector into that same SEZ frame.
        5. Compute the range as the norm of the difference between these two position vectors.
        6. Compute the elevation, and ensure it is between +/- 180 degrees.
        7. Compute the azimuth with a quadrant check, and ensure it is between 0 and 360 degrees."""

    def azimuth_elevation_range_sez_from_location(
        self,
        rx: astro.Orbit,
        location: astro.Location,
        obstructing_body: typing.Optional[astro.Frame] = None,
        ab_corr: typing.Optional[Aberration] = None,
    ) -> astro.AzElRange:
        """Computes the azimuth (in degrees), elevation (in degrees), and range (in kilometers) of the
        receiver state (`rx`) seen from the provided location (as transmitter state, once converted into the SEZ frame of the transmitter.
        Refer to [azimuth_elevation_range_sez] for algorithm details.
        Location terrain masks are always applied, i.e. if the terrain masks the object, all data is set to f64::NAN, unless specified otherwise in the Location."""

    def azimuth_elevation_range_sez_from_location_id(
        self,
        rx: astro.Orbit,
        location_id: int,
        obstructing_body: typing.Optional[astro.Frame] = None,
        ab_corr: typing.Optional[Aberration] = None,
    ) -> astro.AzElRange:
        """Computes the azimuth (in degrees), elevation (in degrees), and range (in kilometers) of the
        receiver state (`rx`) seen from the location ID (as transmitter state, once converted into the SEZ frame of the transmitter.
        Refer to [azimuth_elevation_range_sez] for algorithm details."""

    def azimuth_elevation_range_sez_from_location_name(
        self,
        rx: astro.Orbit,
        location_name: str,
        obstructing_body: typing.Optional[astro.Frame] = None,
        ab_corr: typing.Optional[Aberration] = None,
    ) -> astro.AzElRange:
        """Computes the azimuth (in degrees), elevation (in degrees), and range (in kilometers) of the
        receiver state (`rx`) seen from the location ID (as transmitter state, once converted into the SEZ frame of the transmitter.
        Refer to [azimuth_elevation_range_sez] for algorithm details."""

    def azimuth_elevation_range_sez_many(
        self,
        rx_tx_states: typing.List[astro.Orbit],
        obstructing_body: typing.Optional[astro.Frame] = None,
        ab_corr: typing.Optional[Aberration] = None,
    ) -> typing.List[astro.AzElRange]:
        """Computes the azimuth (in degrees), elevation (in degrees), and range (in kilometers) of the
        receiver states (first item in tuple) seen from the transmitter state (second item in states tuple), once converted into the SEZ frame of the transmitter.

        Note: if any computation fails, the error will be printed to the stderr.
        Note: the output AER will be chronologically sorted, regardless of transmitter.

        Refer to [azimuth_elevation_range_sez] for details."""

    def beta_angle_deg(
        self, state: astro.Orbit, ab_corr: typing.Optional[Aberration] = None
    ) -> float:
        """Computes the Beta angle (β) for a given orbital state, in degrees. A Beta angle of 0° indicates that the orbit plane is edge-on to the Sun, leading to maximum eclipse time. Conversely, a Beta angle of +90° or -90° means the orbit plane is face-on to the Sun, resulting in continuous sunlight exposure and no eclipses.

        The Beta angle (β) is defined as the angle between the orbit plane of a spacecraft and the vector from the central body (e.g., Earth) to the Sun. In simpler terms, it measures how much of the time a satellite in orbit is exposed to direct sunlight.
        The mathematical formula for the Beta angle is: β=arcsin(h⋅usun\u200b)
        Where:
        - h is the unit vector of the orbital momentum.
        - usun\u200b is the unit vector pointing from the central body to the Sun.

        Original code from GMAT, <https://github.com/ChristopherRabotin/GMAT/blob/GMAT-R2022a/src/gmatutil/util/CalculationUtilities.cpp#L209-L219>"""

    def bpc_domain(self, id: int) -> typing.Tuple:
        """Returns the applicable domain of the request id, i.e. start and end epoch that the provided id has loaded data."""

    def bpc_domains(self) -> typing.Dict:
        """Returns a map of each loaded BPC ID to its domain validity.

        # Warning
        This function performs a memory allocation."""

    def bpc_summaries(self, id: int) -> typing.List:
        """Returns a vector of the summaries whose ID matches the desired `id`, in the order in which they will be used, i.e. in reverse loading order.

        # Warning
        This function performs a memory allocation."""

    def bpc_swap(self, alias: str, new_bpc_path: str, new_alias: str) -> None:
        """Load a new DAF/BPC file in place of the one in the provided alias.

        This reuses the existing memory buffer, growing it only if the new file
        is larger than the previous capacity. This effectively adopts a
        "high watermark" memory strategy, where the memory usage for this slot
        is determined by the largest file ever loaded into it."""

    def bpc_unload(self, alias: str) -> None:
        """Unloads (in-place) the BPC with the provided alias.
        **WARNING:** This causes the order of the loaded files to be perturbed, which may be an issue if several SPKs with the same IDs are loaded."""

    def describe(
        self,
        spk: typing.Optional[bool] = None,
        bpc: typing.Optional[bool] = None,
        planetary: typing.Optional[bool] = None,
        spacecraft: typing.Optional[bool] = None,
        eulerparams: typing.Optional[bool] = None,
        locations: typing.Optional[bool] = None,
        time_scale: typing.Optional[time.TimeScale] = None,
        round_time: typing.Optional[bool] = None,
    ) -> None:
        """Pretty prints the description of this Almanac, showing everything by default. Default time scale is TDB.
        If any parameter is set to true, then nothing other than that will be printed."""

    def frame_info(self, uid: astro.Frame) -> astro.Frame:
        """Returns the frame information (gravitational param, shape) as defined in this Almanac from an empty frame"""

    @staticmethod
    def from_ccsds_oem_file(path: str, naif_id: int) -> Almanac:
        """Initializes a new Almanac from a file path to CCSDS OEM file, after converting to to SPICE SPK/BSP"""

    def line_of_sight_obstructed(
        self,
        observer: astro.Orbit,
        observed: astro.Orbit,
        obstructing_body: astro.Frame,
        ab_corr: typing.Optional[Aberration] = None,
    ) -> bool:
        """Computes whether the line of sight between an observer and an observed Cartesian state is obstructed by the obstructing body.
        Returns true if the obstructing body is in the way, false otherwise.

        For example, if the Moon is in between a Lunar orbiter (observed) and a ground station (observer), then this function returns `true`
        because the Moon (obstructing body) is indeed obstructing the line of sight.

        ```text
        Observed
        o  -
        +    -
        +      -
        + ***   -
        * +    *   -
        *  + + * + + o
        *     *     Observer
        ****
        ```

        Key Elements:
        - `o` represents the positions of the observer and observed objects.
        - The dashed line connecting the observer and observed is the line of sight.

        Algorithm (source: Algorithm 35 of Vallado, 4th edition, page 308.):
        - `r1` and `r2` are the transformed radii of the observed and observer objects, respectively.
        - `r1sq` and `r2sq` are the squared magnitudes of these vectors.
        - `r1dotr2` is the dot product of `r1` and `r2`.
        - `tau` is a parameter that determines the intersection point along the line of sight.
        - The condition `(1.0 - tau) * r1sq + r1dotr2 * tau <= ob_mean_eq_radius_km^2` checks if the line of sight is within the obstructing body's radius, indicating an obstruction."""

    def list_kernels(
        self,
        spk: typing.Optional[bool] = None,
        bpc: typing.Optional[bool] = None,
        planetary: typing.Optional[bool] = None,
        spacecraft: typing.Optional[bool] = None,
        eulerparams: typing.Optional[bool] = None,
        locations: typing.Optional[bool] = None,
    ) -> list:
        """Returns the list of loaded kernels"""

    def load(self, path: str) -> Almanac:
        """Generic function that tries to load the provided path guessing to the file type."""

    def load_ccsds_oem_file(self, path: str, naif_id: int) -> Almanac:
        """Converts the provided CCSDS OEM to SPICE SPK/BSP and loads it in the Almanac."""

    def load_from_metafile(self, metafile: MetaFile, autodelete: bool) -> Almanac:
        """Load from the provided MetaFile, downloading it if necessary.
        Set autodelete to true to automatically delete lock files. Lock files are important in multi-threaded loads."""

    def load_stk_e_file(self, path: str, naif_id: int) -> Almanac:
        """Converts the provided Ansys STK .e file to SPICE SPK/BSP and loads it in the Almanac."""

    def location_from_id(self, id: int) -> astro.Location:
        """Returns the Location from its ID, searching through all loaded location datasets in reverse order."""

    def location_from_name(self, name: str) -> astro.Location:
        """Returns the Location from its name, searching through all loaded location datasets in reverse order."""

    def occultation(
        self,
        back_frame: astro.Frame,
        front_frame: astro.Frame,
        observer: astro.Orbit,
        ab_corr: typing.Optional[Aberration] = None,
    ) -> astro.Occultation:
        """Computes the occultation percentage of the `back_frame` object by the `front_frame` object as seen from the observer, when according for the provided aberration correction.

        A zero percent occultation means that the back object is fully visible from the observer.
        A 100%  percent occultation means that the back object is fully hidden from the observer because of the front frame (i.e. _umbra_ if the back object is the Sun).
        A value in between means that the back object is partially hidden from the observser (i.e. _penumbra_ if the back object is the Sun).
        Refer to the [MathSpec](https://nyxspace.com/nyxspace/MathSpec/celestial/eclipse/) for modeling details."""

    def report_event_arcs(
        self,
        state_spec: analysis.StateSpec,
        event: analysis.Event,
        start_epoch: time.Epoch,
        end_epoch: time.Epoch,
    ) -> list:
        """Report the rising and falling edges/states where the event arc happens.

        For example, for a scalar expression less than X, this will report all of the times when the expression falls below X and rises above X.
        This method uses the report_events function under the hood."""

    def report_events(
        self,
        state_spec: analysis.StateSpec,
        event: analysis.Event,
        start_epoch: time.Epoch,
        end_epoch: time.Epoch,
    ) -> list:
        """Report all of the states when the provided event happens.
        This method may only be used for equality events, minimum, and maximum events. For spanned events (e.g. Less Than/Greater Than), use report_event_arcs.

        # Method
        The report event function starts by lineraly scanning the whole state spec from the start to the end epoch.
        This uses an adaptive step scan modeled on the Runge Kutta adaptive step integrator, but the objective is to ensure that the scalar expression
        of the event is evaluated at steps where it is linearly changing (to within 10% of linearity). This allows finding coarse brackets where
        the expression changes signs exactly once.
        Then, each bracket it sent in parallel to a Brent's method root finder to find the exact time of the event.

        # Limitation
        While this approach is both very robust and very fast, if you think the finder may be missing some events, you should _reduce_ the epoch precision
        of the event as a multiplicative factor of that precision is used to scan the trajectory linearly. Alternatively, you may export the scalars at
        a fixed interval using the report_scalars or report_scalars_flat function and manually analyze the results of the scalar expression."""

    def report_scalars(
        self, report: analysis.ReportScalars, time_series: time.TimeSeries
    ) -> dict:
        """Report a set of scalar expressions, optionally with aliases, at a fixed time step defined in the TimeSeries."""

    def report_visibility_arcs(
        self,
        state_spec: analysis.StateSpec,
        location_id: int,
        start_epoch: time.Epoch,
        end_epoch: time.Epoch,
        sample_rate: time.Duration,
        obstructing_body: typing.Optional[astro.Frame] = None,
    ) -> list:
        """Report the list of visibility arcs for the desired location ID."""

    def rotate(
        self, from_frame: astro.Frame, to_frame: astro.Frame, epoch: time.Epoch
    ) -> rotation.DCM:
        """Returns the 6x6 DCM needed to rotation the `from_frame` to the `to_frame`.

        # Warning
        This function only performs the rotation and no translation whatsoever. Use the `transform_from_to` function instead to include rotations.

        # Note
        This function performs a recursion of no more than twice the MAX_TREE_DEPTH."""

    def rotate_to(self, state: astro.Orbit, observer_frame: astro.Frame) -> astro.Orbit:
        """Rotates the provided Cartesian state into the requested observer frame

        **WARNING:** This function only performs the translation and no rotation _whatsoever_. Use the `transform_to` function instead to include rotations."""

    def solar_eclipsing(
        self,
        eclipsing_frame: astro.Frame,
        observer: astro.Orbit,
        ab_corr: typing.Optional[Aberration] = None,
    ) -> astro.Occultation:
        """Computes the solar eclipsing of the observer due to the eclipsing_frame.

        This function calls `occultation` where the back object is the Sun in the J2000 frame, and the front object
        is the provided eclipsing frame."""

    def solar_eclipsing_many(
        self,
        eclipsing_frame: astro.Frame,
        observers: typing.List[astro.Orbit],
        ab_corr: typing.Optional[Aberration] = None,
    ) -> typing.List[astro.Occultation]:
        """Computes the solar eclipsing of all the observers due to the eclipsing_frame, computed in parallel under the hood.

        Note: if any computation fails, the error will be printed to the stderr.
        Note: the output AER will be chronologically sorted, regardless of transmitter.

        Refer to [solar_eclipsing] for details."""

    def spk_domain(self, id: int) -> typing.Tuple:
        """Returns the applicable domain of the request id, i.e. start and end epoch that the provided id has loaded data."""

    def spk_domains(self) -> typing.Dict:
        """Returns a map of each loaded SPK ID to its domain validity.

        # Warning
        This function performs a memory allocation."""

    def spk_ezr(
        self,
        target: int,
        epoch: time.Epoch,
        frame: int,
        observer: int,
        ab_corr: typing.Optional[Aberration] = None,
    ) -> astro.Orbit:
        """Alias fo SPICE's `spkezr` where the inputs must be the NAIF IDs of the objects and frames with the caveat that the aberration is moved to the last positional argument."""

    def spk_summaries(self, id: int) -> typing.List:
        """Returns a vector of the summaries whose ID matches the desired `id`, in the order in which they will be used, i.e. in reverse loading order.

        # Warning
        This function performs a memory allocation."""

    def spk_swap(self, alias: str, new_spk_path: str, new_alias: str) -> None:
        """Load a new DAF/SPK file in place of the one in the provided alias.

        This reuses the existing memory buffer, growing it only if the new file
        is larger than the previous capacity. This effectively adopts a
        "high watermark" memory strategy, where the memory usage for this slot
        is determined by the largest file ever loaded into it
        ."""

    def spk_unload(self, alias: str) -> None:
        """Unloads (in-place) the SPK with the provided alias.
        **WARNING:** This causes the order of the loaded files to be perturbed, which may be an issue if several SPKs with the same IDs are loaded."""

    def state_of(
        self,
        object_id: int,
        observer: astro.Frame,
        epoch: time.Epoch,
        ab_corr: typing.Optional[Aberration] = None,
    ) -> astro.Orbit:
        """Returns the Cartesian state of the object as seen from the provided observer frame (essentially `spkezr`).

        # Note
        The units will be those of the underlying ephemeris data (typically km and km/s)"""

    def sun_angle_deg(
        self, target_id: int, observer_id: int, epoch: time.Epoch, ab_corr: Aberration
    ) -> float:
        """Returns the angular separation (between 0 and 180 degrees) between the observer and the Sun, and the observer and the target body ID.
This is formally known as the "solar elongation".
This computes the Sun Probe Earth angle (SPE) if the probe is in a loaded SPK, its ID is the "observer_id", and the target is set to its central body.

# Geometry
If the SPE is greater than 90 degrees, then the celestial object below the probe is in sunlight.

This angle determines the illumination phase of the target as seen by the observer:
* **~0° (Conjunction):** The Target is in the same direction as the Sun. The observer sees the unlit side ("New Moon").
* **~180° (Opposition):** The Target is in the opposite direction of the Sun. The observer sees the fully lit side ("Full Moon").
* **> 90°:** The observer is generally on the "day" side of the target.


## Sunrise at nadir
```text
Sun
|  \\
|   \\
|    \\
Obs. -- Target
```
## Sun high at nadir
```text
Sun
\\
\\  __ θ > 90
\\     \\
Obs. ---------- Target
```

## Sunset at nadir
```text
Sun
/
/  __ θ < 90
/    /
Obs. -- Target
```

# Algorithm
1. Compute the position of the Sun as seen from the observer
2. Compute the position of the target as seen from the observer
3. Return the arccosine of the dot product of the norms of these vectors."""

    def sun_angle_deg_from_frame(
        self,
        target: astro.Frame,
        observer: astro.Frame,
        epoch: time.Epoch,
        ab_corr: Aberration,
    ) -> float:
        """Convenience function that calls `sun_angle_deg` with the provided frames instead of the ephemeris ID."""

    def to_metaalmanac(self) -> MetaAlmanac:
        """Saves the current configuration to a MetaAlmanac for future reloading from the local file system.

        WARNING: If data was loaded from its raw bytes, or if a custom alias was used, then the MetaFile produced will not be usable.
        The alias used for each data type is expected to be a path. Further, all paths are ASSUMED to be loaded from the same directory.
        The Almanac does not resolve directories for you."""

    def transform(
        self,
        target_frame: astro.Frame,
        observer_frame: astro.Frame,
        epoch: time.Epoch,
        ab_corr: typing.Optional[Aberration] = None,
    ) -> astro.Orbit:
        """Returns the Cartesian state needed to transform the `from_frame` to the `to_frame`.

        # SPICE Compatibility
        This function is the SPICE equivalent of spkezr: `spkezr(TARGET_ID, EPOCH_TDB_S, ORIENTATION_ID, ABERRATION, OBSERVER_ID)`
        In ANISE, the TARGET_ID and ORIENTATION are provided in the first argument (TARGET_FRAME), as that frame includes BOTH
        the target ID and the orientation of that target. The EPOCH_TDB_S is the epoch in the TDB time system, which is computed
        in ANISE using Hifitime. THe ABERRATION is computed by providing the optional Aberration flag. Finally, the OBSERVER
        argument is replaced by OBSERVER_FRAME: if the OBSERVER_FRAME argument has the same orientation as the TARGET_FRAME, then this call
        will return exactly the same data as the spkerz SPICE call.

        # Note
        The units will be those of the underlying ephemeris data (typically km and km/s)"""

    def transform_many(
        self,
        target_frame: astro.Frame,
        observer_frame: astro.Frame,
        time_series: time.TimeSeries,
        ab_corr: typing.Optional[Aberration] = None,
    ) -> typing.List[astro.Orbit]:
        """Returns a chronologically sorted list of the Cartesian states that transform the `from_frame` to the `to_frame` for each epoch of the time series, computed in parallel under the hood.
        Note: if any transformation fails, the error will be printed to the stderr.

        Refer to [transform] for details."""

    def transform_many_to(
        self,
        states: typing.List[astro.Orbit],
        observer_frame: astro.Frame,
        ab_corr: typing.Optional[Aberration] = None,
    ) -> typing.List[astro.Orbit]:
        """Returns a chronologically sorted list of the provided states as seen from the observer frame, given the aberration.
        Note: if any transformation fails, the error will be printed to the stderr.
        Note: the input ordering is lost: the output states will not be in the same order as the input states if these are not chronologically sorted!

        Refer to [transform_to] for details."""

    def transform_to(
        self,
        state: astro.Orbit,
        observer_frame: astro.Frame,
        ab_corr: typing.Optional[Aberration] = None,
    ) -> astro.Orbit:
        """Returns the provided state as seen from the observer frame, given the aberration."""

    def translate(
        self,
        target_frame: astro.Frame,
        observer_frame: astro.Frame,
        epoch: time.Epoch,
        ab_corr: typing.Optional[Aberration] = None,
    ) -> astro.Orbit:
        """Returns the Cartesian state of the target frame as seen from the observer frame at the provided epoch, and optionally given the aberration correction.

        # SPICE Compatibility
        This function is the SPICE equivalent of spkezr: `spkezr(TARGET_ID, EPOCH_TDB_S, ORIENTATION_ID, ABERRATION, OBSERVER_ID)`
        In ANISE, the TARGET_ID and ORIENTATION are provided in the first argument (TARGET_FRAME), as that frame includes BOTH
        the target ID and the orientation of that target. The EPOCH_TDB_S is the epoch in the TDB time system, which is computed
        in ANISE using Hifitime. THe ABERRATION is computed by providing the optional Aberration flag. Finally, the OBSERVER
        argument is replaced by OBSERVER_FRAME: if the OBSERVER_FRAME argument has the same orientation as the TARGET_FRAME, then this call
        will return exactly the same data as the spkerz SPICE call.

        # Warning
        This function only performs the translation and no rotation whatsoever. Use the `transform` function instead to include rotations.

        # Note
        This function performs a recursion of no more than twice the [MAX_TREE_DEPTH]."""

    def translate_geometric(
        self, target_frame: astro.Frame, observer_frame: astro.Frame, epoch: time.Epoch
    ) -> astro.Orbit:
        """Returns the geometric position vector, velocity vector, and acceleration vector needed to translate the `from_frame` to the `to_frame`, where the distance is in km, the velocity in km/s, and the acceleration in km/s^2."""

    def translate_to(
        self,
        state: astro.Orbit,
        observer_frame: astro.Frame,
        ab_corr: typing.Optional[Aberration] = None,
    ) -> astro.Orbit:
        """Translates the provided Cartesian state into the requested observer frame

        **WARNING:** This function only performs the translation and no rotation _whatsoever_. Use the `transform_to` function instead to include rotations."""

    def translate_to_parent(
        self, source: astro.Frame, epoch: time.Epoch
    ) -> astro.Orbit:
        """Performs the GEOMETRIC translation to the parent. Use translate_from_to for aberration."""

    def __repr__(self) -> str:
        """Return repr(self)."""

    def __str__(self) -> str:
        """Return str(self)."""

@typing.final
class LocationDataSet:
    """A wrapper around a location dataset kernel (PyO3 does not handle type aliases).
    Use this class to load and unload kernels. Manipulate using its LocationDhallSet representation."""

    def __init__(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        """Initialize self.  See help(type(self)) for accurate signature."""

    def __new__(cls) -> LocationDataSet:
        """A wrapper around a location dataset kernel (PyO3 does not handle type aliases).
        Use this class to load and unload kernels. Manipulate using its LocationDhallSet representation."""

    @staticmethod
    def load(path: str) -> LocationDataSet:
        """Loads a Location Dataset kernel from the provided path"""

    def save_as(self, path: str, overwrite: typing.Optional[bool] = False) -> None:
        """Save this dataset as a kernel, optionally specifying whether to overwrite the existing file."""

    def to_dhallset(self) -> LocationDhallSet:
        """Converts this location dataset into a manipulable location Dhall set."""

@typing.final
class LocationDhallSet:
    """A Dhall-serializable Location DataSet that serves as an optional intermediate to the LocationDataSet kernels."""

    data: list

    def __init__(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        """Initialize self.  See help(type(self)) for accurate signature."""

    def __new__(cls, data: list) -> LocationDhallSet:
        """A Dhall-serializable Location DataSet that serves as an optional intermediate to the LocationDataSet kernels."""

    def dumps(self) -> str:
        """Returns the Dhall representation of this LocationDhallSet. Equivalent to to_dhall."""

    @staticmethod
    def from_dhall(repr: str) -> LocationDhallSet:
        """Loads this Location dataset from its Dhall representation as a string"""

    @staticmethod
    def loads(repr: str) -> LocationDhallSet:
        """Loads this Location dataset from its Dhall representation as a string. Equivalent to from_dhall."""

    def to_dataset(self) -> LocationDataSet:
        """Converts this location Dhall set into a Python-compatible Location DataSet."""

    def to_dhall(self) -> str:
        """Returns the Dhall representation of this Location"""

@typing.final
class LocationDhallSetEntry:
    """Entry of a Location Dhall set"""

    alias: str
    id: int
    value: astro.Location

    def __init__(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        """Initialize self.  See help(type(self)) for accurate signature."""

    def __new__(
        cls,
        value: astro.Location,
        id: typing.Optional[int] = None,
        alias: typing.Optional[str] = None,
    ) -> LocationDhallSetEntry:
        """Entry of a Location Dhall set"""

@typing.final
class MetaAlmanac:
    """A structure to set up an Almanac, with automatic downloading, local storage, checksum checking, and more.

    # Behavior
    If the URI is a local path, relative or absolute, nothing will be fetched from a remote. Relative paths are relative to the execution folder (i.e. the current working directory).
    If the URI is a remote path, the MetaAlmanac will first check if the file exists locally. If it exists, it will check that the CRC32 checksum of this file matches that of the specs.
    If it does not match, the file will be downloaded again. If no CRC32 is provided but the file exists, then the MetaAlmanac will fetch the remote file and overwrite the existing file.
    The downloaded path will be stored in the "AppData" folder."""

    files: typing.List

    def __init__(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        """Initialize self.  See help(type(self)) for accurate signature."""

    def __new__(cls, maybe_path: typing.Optional[str] = None) -> MetaAlmanac:
        """A structure to set up an Almanac, with automatic downloading, local storage, checksum checking, and more.

        # Behavior
        If the URI is a local path, relative or absolute, nothing will be fetched from a remote. Relative paths are relative to the execution folder (i.e. the current working directory).
        If the URI is a remote path, the MetaAlmanac will first check if the file exists locally. If it exists, it will check that the CRC32 checksum of this file matches that of the specs.
        If it does not match, the file will be downloaded again. If no CRC32 is provided but the file exists, then the MetaAlmanac will fetch the remote file and overwrite the existing file.
        The downloaded path will be stored in the "AppData" folder."""

    def dumps(self) -> str:
        """Dumps the configured Meta Almanac into a Dhall string. Equivalent to to_dhall()."""

    @staticmethod
    def from_dhall(repr: str) -> MetaAlmanac:
        """Loads this Meta Almanac from its Dhall string representation"""

    @staticmethod
    def latest(autodelete: typing.Optional[bool] = None) -> Almanac:
        """Returns an Almanac loaded from the latest NAIF data via the `default` MetaAlmanac.
        The MetaAlmanac will download the DE440s.bsp file, the PCK0008.PCA, the full Moon Principal Axis BPC (moon_pa_de440_200625) and the latest high precision Earth kernel from JPL.

        # File list
        - <http://public-data.nyxspace.com/anise/de440s.bsp>
        - <http://public-data.nyxspace.com/anise/v0.5/pck08.pca>
        - <http://public-data.nyxspace.com/anise/moon_pa_de440_200625.bpc>
        - <https://naif.jpl.nasa.gov/pub/naif/generic_kernels/pck/earth_latest_high_prec.bpc>

        # Reproducibility

        Note that the `earth_latest_high_prec.bpc` file is regularly updated daily (or so). As such,
        if queried at some future time, the Earth rotation parameters may have changed between two queries.

        Set `autodelete` to true to delete lock file if a dead lock is detected after 10 seconds."""

    @staticmethod
    def loads(s: str) -> MetaAlmanac:
        """Loads the provided string as a Dhall configuration to build a MetaAlmanac"""

    def process(self, autodelete: typing.Optional[bool] = None) -> Almanac:
        """Fetch all of the URIs and return a loaded Almanac.
        When downloading the data, ANISE will create a temporarily lock file to prevent race conditions
        where multiple processes download the data at the same time. Set `autodelete` to true to delete
        this lock file if a dead lock is detected after 10 seconds. Set this flag to false if you have
        more than ten processes which may attempt to download files in parallel."""

    def to_dhall(self) -> str:
        """Serializes the configurated Meta Almanac into a Dhall string. Equivalent to dumps()."""

    def __eq__(self, value: typing.Any) -> bool:
        """Return self==value."""

    def __ge__(self, value: typing.Any) -> bool:
        """Return self>=value."""

    def __gt__(self, value: typing.Any) -> bool:
        """Return self>value."""

    def __le__(self, value: typing.Any) -> bool:
        """Return self<=value."""

    def __lt__(self, value: typing.Any) -> bool:
        """Return self<value."""

    def __ne__(self, value: typing.Any) -> bool:
        """Return self!=value."""

    def __repr__(self) -> str:
        """Return repr(self)."""

    def __str__(self) -> str:
        """Return str(self)."""

@typing.final
class MetaFile:
    """MetaFile allows downloading a remote file from a URL (http, https only), and interpolation of paths in environment variable using the Dhall syntax `env:MY_ENV_VAR`.

    The data is stored in the user's local temp directory (i.e. `~/.local/share/nyx-space/anise/` on Linux and `AppData/Local/nyx-space/anise/` on Windows).
    Prior to loading a remote resource, if the local resource exists, its CRC32 will be computed: if it matches the CRC32 of this instance of MetaFile,
    then the file will not be downloaded a second time."""

    crc32: int
    uri: str

    def __init__(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        """Initialize self.  See help(type(self)) for accurate signature."""

    def __new__(cls, uri: str, crc32: typing.Optional[int] = None) -> MetaFile:
        """MetaFile allows downloading a remote file from a URL (http, https only), and interpolation of paths in environment variable using the Dhall syntax `env:MY_ENV_VAR`.

        The data is stored in the user's local temp directory (i.e. `~/.local/share/nyx-space/anise/` on Linux and `AppData/Local/nyx-space/anise/` on Windows).
        Prior to loading a remote resource, if the local resource exists, its CRC32 will be computed: if it matches the CRC32 of this instance of MetaFile,
        then the file will not be downloaded a second time."""

    def process(self, autodelete: typing.Optional[bool] = None) -> None:
        """Processes this MetaFile by downloading it if it's a URL.

        This function modified `self` and changes the URI to be the path to the downloaded file."""

    def __eq__(self, value: typing.Any) -> bool:
        """Return self==value."""

    def __ge__(self, value: typing.Any) -> bool:
        """Return self>=value."""

    def __gt__(self, value: typing.Any) -> bool:
        """Return self>value."""

    def __le__(self, value: typing.Any) -> bool:
        """Return self<=value."""

    def __lt__(self, value: typing.Any) -> bool:
        """Return self<value."""

    def __ne__(self, value: typing.Any) -> bool:
        """Return self!=value."""

    def __repr__(self) -> str:
        """Return repr(self)."""

    def __str__(self) -> str:
        """Return str(self)."""

@typing.final
class PyReportScalars:
    """A basic report builder that can be serialized seperately from the execution.
    The scalars must be a tuple of (ScalarExpr, String) where the String is the alias (optional)."""

    def __init__(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        """Initialize self.  See help(type(self)) for accurate signature."""

    def __new__(cls, scalars: list, state_spec: analysis.StateSpec) -> PyReportScalars:
        """A basic report builder that can be serialized seperately from the execution.
        The scalars must be a tuple of (ScalarExpr, String) where the String is the alias (optional)."""

    @staticmethod
    def from_s_expr(expr: str) -> analysis.ReportScalars:
        """Convert the S-Expression to a report builder"""

    def to_s_expr(self) -> str:
        """Converts this report builder to its S-Expression"""

def exec_gui(): ...

__author__: str = "Christopher Rabotin <christopher.rabotin@gmail.com>"
__version__: str = "0.9.5"



from __future__ import annotations
import typing

def convert_fk(fk_file_path: str, anise_output_path: str, show_comments: typing.Optional[bool]=None, overwrite: typing.Optional[bool]=None) -> None:
    """Converts a KPL/FK file, that defines frame constants like fixed rotations, and frame name to ID mappings into the EulerParameterDataSet equivalent ANISE file.
KPL/FK files must be converted into "PCA" (Planetary Constant ANISE) files before being loaded into ANISE."""

def convert_tpc(pck_file_path: str, gm_file_path: str, anise_output_path: str, overwrite: typing.Optional[bool]=None) -> None:
    """Converts two KPL/TPC files, one defining the planetary constants as text, and the other defining the gravity parameters, into the PlanetaryDataSet equivalent ANISE file.
KPL/TPC files must be converted into "PCA" (Planetary Constant ANISE) files before being loaded into ANISE."""


from anise.time import *

"""
The time tests only make sure that we can call all of the functions that are re-exported.
For comprehensive tests of the time, refer to the hifitime test suite
"""


def test_exports():
    for cls in [
        Epoch,
        TimeSeries,
        Duration,
        Unit,
        Ut1Provider,
        LatestLeapSeconds,
        LeapSecondsFile,
    ]:
        print(f"{cls} OK")


if __name__ == "__main__":
    test_exports()
